import contextlib
import copy
import datetime
from collections import Counter
from types import TracebackType
from typing import Any, Callable, Generator, MutableMapping, Optional, Type
from xml.dom import minidom  # noqa: S408

import pytest
from structlog import _config as structlog_config
from structlog import contextvars as structlog_contextvars
from structlog.testing import LogCapture
from structlog.types import EventDict

from ffun.core.postgresql import execute
from ffun.domain.entities import UserId

PRODUCER = Callable[[], Any]


class Comparator:
    message = "Error"

    def __init__(
        self,
        producer: PRODUCER,
        message: Optional[str] = None,
        template: str = "{message}: original={old_value}, new={new_value}.",
    ):
        if message is not None:
            self.message = message

        if not callable(producer):
            self.producer = lambda: copy.deepcopy(producer)
        else:
            self.producer = producer  # type: ignore

        self.old_value: Any = None
        self.template = template

    def __enter__(self) -> "Comparator":
        self.old_value = self.producer()  # type: ignore
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        __tracebackhide__ = True  # pylint: disable=W0612

        new_value = self.producer()  # type: ignore

        if self.validate(self.old_value, new_value):
            self.fail(self.old_value, new_value)

    async def __aenter__(self) -> "Comparator":
        self.old_value = await self.producer()  # type: ignore
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        __tracebackhide__ = True  # pylint: disable=W0612

        new_value = await self.producer()  # type: ignore

        if self.validate(self.old_value, new_value):
            self.fail(self.old_value, new_value)

    def fail(self, old_value: Any, new_value: Any) -> None:
        pytest.fail(self.template.format(message=self.message, old_value=old_value, new_value=new_value))

    def validate(self, old_value: Any, new_value: Any) -> bool:
        raise NotImplementedError()


class NotChanged(Comparator):
    message = "Value changed"

    def validate(self, old_value: Any, new_value: Any) -> bool:
        return bool(old_value != new_value)


class Changed(Comparator):
    message = "Value has not changed"

    def validate(self, old_value: Any, new_value: Any) -> bool:
        return bool(old_value == new_value)


class Increased(Comparator):
    message = "Value has not increased"

    def validate(self, old_value: Any, new_value: Any) -> bool:
        return bool(old_value < new_value)


class Decreased(Comparator):
    message = "Value has not decreased"

    def validate(self, old_value: Any, new_value: Any) -> bool:
        return bool(old_value < new_value)


class Delta(Comparator):
    def __init__(self, producer: PRODUCER, delta: Any, **kwargs: Any):
        if "message" not in kwargs:
            kwargs["message"] = f"Value has not changed to delta {delta}"

        super().__init__(producer, **kwargs)

        self.delta = delta

    def validate(self, old_value: Any, new_value: Any) -> bool:
        return bool(new_value - old_value != self.delta)


class TableSizeMixin:
    async def _producer(self) -> int:
        results = await execute(f"SELECT count(*) AS number FROM {self.table}")  # type: ignore # noqa: S608
        size = results[0]["number"]
        assert isinstance(size, int)
        return size


class TableSizeDelta(Delta, TableSizeMixin):
    def __init__(self, table: str, **kwargs: Any) -> None:
        if "producer" not in kwargs:
            kwargs["producer"] = self._producer

        super().__init__(**kwargs)

        self.table = table


class TableSizeNotChanged(NotChanged, TableSizeMixin):
    def __init__(self, table: str, **kwargs: Any) -> None:
        if "producer" not in kwargs:
            kwargs["producer"] = self._producer

        super().__init__(**kwargs)

        self.table = table


class TableSizeDecreased(Decreased, TableSizeMixin):
    def __init__(self, table: str, **kwargs: Any) -> None:
        if "producer" not in kwargs:
            kwargs["producer"] = self._producer

        super().__init__(**kwargs)

        self.table = table


class TableSizeIncreased(Increased, TableSizeMixin):
    def __init__(self, table: str, **kwargs: Any) -> None:
        if "producer" not in kwargs:
            kwargs["producer"] = self._producer

        super().__init__(**kwargs)

        self.table = table


def assert_times_is_near(
    a: datetime.datetime, b: datetime.datetime, delta: datetime.timedelta = datetime.timedelta(seconds=1)
) -> None:
    assert abs(a - b) < delta


def assert_logs(logs: list[MutableMapping[str, Any]], **kwargs: int) -> None:
    found_enents = Counter(log["event"] for log in logs)

    for key, expected_count in kwargs.items():
        if expected_count != found_enents.get(key, 0):
            pytest.fail(
                f"Key {key} not found {expected_count} times in logs, but found {found_enents.get(key, 0)} times"
            )


def assert_log_context_vars(**expected: Any) -> None:
    bound_vars = structlog_contextvars.get_contextvars()

    for key, value in expected.items():
        assert bound_vars.get(key) == value, f"Key {key} = {bound_vars} not equal to expected {value}"


def assert_logs_levels(logs: list[MutableMapping[str, Any]], **kwargs: str) -> None:
    for record in logs:
        if record["event"] in kwargs:
            assert record["log_level"] == kwargs[record["event"]], f"Log level is not equal to expected for {record}"


def assert_logs_have_no_errors(logs: list[MutableMapping[str, Any]]) -> None:
    for record in logs:
        if record["log_level"].lower() == "error":
            pytest.fail(f"Error found in logs: {record}")


@contextlib.contextmanager
def capture_logs() -> Generator[list[EventDict], None, None]:
    """
    This is a modified version of capture_logs from structlog.testing

    Differences:

    - merge contextvars
    """
    cap = LogCapture()

    processors = structlog_config.get_config()["processors"]
    old_processors = processors.copy()
    try:
        # clear processors list and use LogCapture for testing
        processors.clear()
        processors.append(structlog_contextvars.merge_contextvars)
        processors.append(cap)
        structlog_config.configure(processors=processors)
        yield cap.entries
    finally:
        # remove LogCapture and restore original processors
        processors.clear()
        processors.extend(old_processors)
        structlog_config.configure(processors=processors)


def assert_logs_has_business_event(  # noqa: CCR001
    logs: list[MutableMapping[str, Any]], name: str, user_id: UserId | None, **atributes: Any
) -> None:

    if user_id is not None:
        user_id = str(user_id)  # type: ignore

    for record in logs:
        if not (record.get("b_kind") == "event" and record["event"] == name and record.get("b_user_id") == user_id):
            continue

        assert "b_uid" in record, "b_uid not found in record"

        for key, value in atributes.items():
            assert key in record["b_attributes"], f"Key {key} not found in record"
            assert (
                record["b_attributes"][key] == value
            ), f"Key {key} = {record["b_attributes"][key]!r} not equal to expected {value!r}"

        break
    else:
        pytest.fail(f"Event {name} not found in logs")


def assert_logs_has_no_business_event(logs: list[MutableMapping[str, Any]], name: str) -> None:
    for record in logs:
        if record.get("b_kind") == "event" and record["event"] == name:
            pytest.fail(f"Event {name} found in logs")


def assert_compare_xml(a: str, b: str) -> None:
    assert minidom.parseString(a).toprettyxml() == minidom.parseString(b).toprettyxml()  # noqa: S318



import pathlib
import uuid

import pytest
import pytest_asyncio

from ffun.domain.domain import new_collection_id, new_feed_id
from ffun.domain.entities import CollectionSlug
from ffun.domain.urls import str_to_absolute_url, to_feed_url, url_to_source_uid
from ffun.feeds import domain as f_domain
from ffun.feeds_collections import errors
from ffun.feeds_collections.collections import Collections
from ffun.feeds_collections.entities import Collection, FeedInfo

_root = pathlib.Path(__file__).parent

_test_collections_configs = _root / "fixtures" / "test_collections"


class TestCollections:

    @pytest_asyncio.fixture(scope="session")
    async def collections(self) -> Collections:
        collections = Collections()

        await collections.initialize(collection_configs=_test_collections_configs)

        return collections

    def test_constructor(self) -> None:
        collections = Collections()

        assert not collections.initialized
        assert collections._collections == []
        assert collections._feeds_in_collections == {}

    @pytest.mark.asyncio
    async def test_initialize__no_feeds(self) -> None:
        collections = Collections()

        await collections.initialize(collection_configs=None)

        assert collections.initialized
        assert collections._collections == []
        assert collections._feeds_in_collections == {}

    # Also tests:
    # - Collections.load
    # - Collections.prepare_feeds
    @pytest.mark.asyncio
    async def test_initialize__has_feeds(self, collections: Collections) -> None:
        assert collections.initialized
        assert {c.name for c in collections._collections} == {"Feeds Fun", "Scientific Papers"}

        assert len(collections._feeds_in_collections) == 3

        feeds = await f_domain.get_feeds(collections._feeds_in_collections)
        feed_ids = {feed.id for feed in feeds}

        expected_urls = set()

        for collection in collections._collections:
            for feed_info in collection.feeds:
                assert feed_info.feed_id in collections._feeds_in_collections
                assert feed_info.feed_id in feed_ids
                expected_urls.add(feed_info.url)

        assert expected_urls == {feed.url for feed in feeds}

        # test that feeds created with correct source ids
        source_uids = {url_to_source_uid(url) for url in expected_urls}
        source_ids = await f_domain.get_source_ids(source_uids)

        for feed in feeds:
            assert feed.source_id == source_ids[url_to_source_uid(feed.url)]

        assert collections.count_total_feeds() == len(feeds)
        assert set(collections.all_feed_ids()) == {feed.id for feed in feeds}

    @pytest.mark.asyncio
    async def test_collections(self, collections: Collections) -> None:
        assert collections.collections() == collections._collections
        assert collections.collections() is not collections._collections

    @pytest.mark.asyncio
    async def test_collection(self, collections: Collections) -> None:
        for collection in collections._collections:
            assert collections.collection(collection.id) == collection

    @pytest.mark.asyncio
    async def test_collection__no_collection(self, collections: Collections) -> None:
        with pytest.raises(errors.CollectionNotFound):
            collections.collection(new_collection_id())

    @pytest.mark.asyncio
    async def test_collection_by_slug(self, collections: Collections) -> None:
        for collection in collections._collections:
            assert collections.collection_by_slug(collection.slug) == collection

    @pytest.mark.asyncio
    async def test_collection_by_slug__no_collection(self, collections: Collections) -> None:
        with pytest.raises(errors.CollectionNotFound):
            collections.collection_by_slug(CollectionSlug(uuid.uuid4().hex))

    @pytest.mark.asyncio
    async def test_has_feed(self, collections: Collections) -> None:
        assert not collections.has_feed(new_feed_id())

        existed_feed_id = list(collections._feeds_in_collections)[0]
        assert collections.has_feed(existed_feed_id)

    @pytest.mark.asyncio
    async def test_get_feed_info__feed_not_found(self, collections: Collections) -> None:
        with pytest.raises(errors.FeedNotFound):
            collections.get_feed_info(new_feed_id())

    @pytest.mark.asyncio
    async def test_get_feed_info(self, collections: Collections) -> None:
        existed_feed_id = list(collections._feeds_in_collections)[0]

        feed_info = collections.get_feed_info(existed_feed_id)

        assert feed_info.feed_id == existed_feed_id

    @pytest.mark.asyncio
    async def test_validate_collection_ids(self) -> None:
        collections = Collections()

        feed_1 = FeedInfo(
            url=to_feed_url(str_to_absolute_url("http://example.com/feed1")), title="Feed 1", description="Feed 1"
        )

        collection_1 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-1"),
            gui_order=1,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        collection_2 = Collection(
            id=collection_1.id,
            slug=CollectionSlug("collection-slug-2"),
            gui_order=2,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        await collections.add_test_collection(collection_1)

        with pytest.raises(errors.DuplicateCollectionIds):
            await collections.add_test_collection(collection_2)

    @pytest.mark.asyncio
    async def test_validate_collection_slugs(self) -> None:
        collections = Collections()

        feed_1 = FeedInfo(
            url=to_feed_url(str_to_absolute_url("http://example.com/feed1")), title="Feed 1", description="Feed 1"
        )

        collection_1 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-1"),
            gui_order=1,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        collection_2 = Collection(
            id=new_collection_id(),
            slug=collection_1.slug,
            gui_order=2,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        await collections.add_test_collection(collection_1)

        with pytest.raises(errors.DuplicateCollectionSlugs):
            await collections.add_test_collection(collection_2)

    @pytest.mark.asyncio
    async def test_validate_gui_order(self) -> None:
        collections = Collections()

        feed_1 = FeedInfo(
            url=to_feed_url(str_to_absolute_url("http://example.com/feed1")), title="Feed 1", description="Feed 1"
        )

        collection_1 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-1"),
            gui_order=1,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        collection_2 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-2"),
            gui_order=1,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        await collections.add_test_collection(collection_1)

        with pytest.raises(errors.DuplicateCollectionOrders):
            await collections.add_test_collection(collection_2)

    @pytest.mark.asyncio
    async def test_validate_emptiness(self) -> None:
        collections = Collections()

        feed_1 = FeedInfo(
            url=to_feed_url(str_to_absolute_url("http://example.com/feed1")), title="Feed 1", description="Feed 1"
        )

        collection_1 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-1"),
            gui_order=1,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[feed_1],
            show_on_main=True,
        )

        collection_2 = Collection(
            id=new_collection_id(),
            slug=CollectionSlug("collection-slug-2"),
            gui_order=2,
            name=uuid.uuid4().hex,
            description=uuid.uuid4().hex,
            feeds=[],
            show_on_main=True,
        )

        await collections.add_test_collection(collection_1)

        with pytest.raises(errors.CollectionIsEmpty):
            await collections.add_test_collection(collection_2)

    def test_collections_for_feed(self, collections: Collections) -> None:
        for collection in collections._collections:
            for feed_info in collection.feeds:
                assert feed_info.feed_id is not None
                assert collection.id in collections.collections_for_feed(feed_info.feed_id)

    def test_collections_for_feed__no_collection(self, collections: Collections) -> None:
        for collection in collections._collections:
            assert collection.id not in collections.collections_for_feed(new_feed_id())



"""
settings-table
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()

sql_create_table = """
CREATE TABLE us_settings (
    user_id UUID NOT NULL,
    kind INTEGER NOT NULL,
    value TEXT NOT NULL,

    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW(),

    PRIMARY KEY (user_id, kind)
)
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_table)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE us_settings")


steps = [step(apply_step, rollback_step)]



"""
add-title-and-description
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__ = {"20230329_01_ilwfq-feeds-table"}


sql = """
ALTER TABLE f_feeds
ADD COLUMN title TEXT DEFAULT NULL,
ADD COLUMN description TEXT DEFAULT NULL
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()

    sql = """
    ALTER TABLE f_feeds
    DROP COLUMN title,
    DROP COLUMN description
    """

    cursor.execute(sql)


steps = [step(apply_step, rollback_step)]



import pytest

from ffun.core.tests.helpers import (
    TableSizeDelta,
    TableSizeNotChanged,
    assert_logs_has_business_event,
    assert_logs_has_no_business_event,
    capture_logs,
)
from ffun.domain.domain import new_rule_id
from ffun.domain.entities import UserId
from ffun.scores import domain, errors, operations


class TestCreateOrUpdateRule:
    @pytest.mark.asyncio
    async def test_create_new_rule(
        self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        required_tags = five_tags_ids[:3]
        excluded_tags = five_tags_ids[3:]

        with capture_logs() as logs:
            async with TableSizeDelta("s_rules", delta=1):
                created_rule = await operations.create_or_update_rule(
                    internal_user_id, score=13, required_tags=required_tags, excluded_tags=excluded_tags
                )

        rules = await domain.get_rules(internal_user_id)

        assert rules == [created_rule]

        assert created_rule.user_id == internal_user_id
        assert created_rule.required_tags == set(required_tags)
        assert created_rule.excluded_tags == set(excluded_tags)
        assert created_rule.score == 13

        assert_logs_has_business_event(
            logs,
            "rule_created",
            user_id=internal_user_id,
            rule_id=str(created_rule.id),
            required_tags=list(required_tags),
            excluded_tags=list(excluded_tags),
            score=13,
        )
        assert_logs_has_no_business_event(logs, "rule_updated")

    @pytest.mark.asyncio
    async def test_tags_order_does_not_affect_creation(
        self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        required_tags = list(five_tags_ids[:3])
        excluded_tags = list(five_tags_ids[3:])

        async with TableSizeDelta("s_rules", delta=1):
            created_rule_1 = await operations.create_or_update_rule(
                internal_user_id, required_tags=required_tags, excluded_tags=excluded_tags, score=13
            )

            created_rule_2 = await operations.create_or_update_rule(
                internal_user_id,
                required_tags=reversed(required_tags),
                excluded_tags=reversed(excluded_tags),
                score=17,
            )

        assert created_rule_1.required_tags == created_rule_2.required_tags
        assert created_rule_1.excluded_tags == created_rule_2.excluded_tags

        rules = await domain.get_rules(internal_user_id)

        assert rules == [created_rule_2]

    @pytest.mark.asyncio
    async def test_update_scores_of_existed_rule(
        self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        required_tags = list(five_tags_ids[:3])
        excluded_tags = list(five_tags_ids[3:])

        await operations.create_or_update_rule(
            internal_user_id, score=13, required_tags=required_tags, excluded_tags=excluded_tags
        )

        with capture_logs() as logs:
            async with TableSizeNotChanged("s_rules"):
                updated_rule = await operations.create_or_update_rule(
                    internal_user_id, score=17, required_tags=required_tags, excluded_tags=excluded_tags
                )

        rules = await domain.get_rules(internal_user_id)

        assert rules == [updated_rule]

        assert updated_rule.user_id == internal_user_id
        assert updated_rule.required_tags == set(required_tags)
        assert updated_rule.excluded_tags == set(excluded_tags)
        assert updated_rule.score == 17

        assert_logs_has_business_event(
            logs,
            "rule_updated",
            user_id=internal_user_id,
            rule_id=str(updated_rule.id),
            required_tags=required_tags,
            excluded_tags=excluded_tags,
            score=17,
        )
        assert_logs_has_no_business_event(logs, "rule_created")

    @pytest.mark.asyncio
    async def test_multiple_entities(
        self, internal_user_id: UserId, another_internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        await operations.create_or_update_rule(
            internal_user_id, required_tags=five_tags_ids[:2], excluded_tags=five_tags_ids[-2:], score=3
        )
        await operations.create_or_update_rule(
            another_internal_user_id, required_tags=five_tags_ids[:2], excluded_tags=five_tags_ids[-2:], score=5
        )
        await operations.create_or_update_rule(
            another_internal_user_id, required_tags=five_tags_ids[1:], excluded_tags=five_tags_ids[:1], score=7
        )
        await operations.create_or_update_rule(
            internal_user_id, required_tags=five_tags_ids[:2], excluded_tags=five_tags_ids[-2:], score=11
        )

        rules = await domain.get_rules(internal_user_id)

        assert len(rules) == 1

        assert rules[0].user_id == internal_user_id
        assert rules[0].required_tags == set(five_tags_ids[:2])
        assert rules[0].excluded_tags == set(five_tags_ids[-2:])
        assert rules[0].score == 11

        rules = await domain.get_rules(another_internal_user_id)

        rules.sort(key=lambda r: r.score)

        assert len(rules) == 2

        assert rules[0].user_id == another_internal_user_id
        assert rules[0].required_tags == set(five_tags_ids[:2])
        assert rules[0].excluded_tags == set(five_tags_ids[-2:])
        assert rules[0].score == 5

        assert rules[1].user_id == another_internal_user_id
        assert rules[1].required_tags == set(five_tags_ids[1:])
        assert rules[1].excluded_tags == set(five_tags_ids[:1])
        assert rules[1].score == 7

    @pytest.mark.asyncio
    async def test_tags_intersection(
        self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        required_tags = five_tags_ids[:3]
        excluded_tags = five_tags_ids[2:]

        assert len(set(required_tags) & set(excluded_tags)) == 1

        with capture_logs() as logs:
            with pytest.raises(errors.TagsIntersection):
                async with TableSizeNotChanged("s_rules"):
                    await operations.create_or_update_rule(
                        internal_user_id, required_tags=required_tags, excluded_tags=excluded_tags, score=13
                    )

        rules = await domain.get_rules(internal_user_id)

        assert rules == []

        assert_logs_has_no_business_event(logs, "rule_created")


class TestDeleteRule:
    @pytest.mark.asyncio
    async def test_delete_rule(
        self, internal_user_id: UserId, another_internal_user_id: UserId, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_delete = await operations.create_or_update_rule(
            internal_user_id, required_tags=three_tags_ids, excluded_tags=[], score=13
        )
        rule_2 = await operations.create_or_update_rule(
            internal_user_id, required_tags=three_tags_ids[:2], excluded_tags=[], score=17
        )
        rule_3 = await operations.create_or_update_rule(
            another_internal_user_id, required_tags=[], excluded_tags=three_tags_ids, score=19
        )

        with capture_logs() as logs:
            async with TableSizeDelta("s_rules", delta=-1):
                await operations.delete_rule(internal_user_id, rule_to_delete.id)

        assert_logs_has_business_event(logs, "rule_deleted", user_id=internal_user_id, rule_id=str(rule_to_delete.id))

        rules = await domain.get_rules(internal_user_id)

        assert rules == [rule_2]

        rules = await domain.get_rules(another_internal_user_id)

        assert rules == [rule_3]

    @pytest.mark.asyncio
    async def test_delete_not_existed_rule(self, internal_user_id: UserId) -> None:
        with capture_logs() as logs:
            async with TableSizeNotChanged("s_rules"):
                await operations.delete_rule(internal_user_id, new_rule_id())

        assert_logs_has_no_business_event(logs, "rule_deleted")

    @pytest.mark.asyncio
    async def test_delete_for_wrong_user(
        self, internal_user_id: UserId, another_internal_user_id: UserId, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_delete = await operations.create_or_update_rule(
            internal_user_id, required_tags=three_tags_ids, excluded_tags=[], score=13
        )

        async with TableSizeNotChanged("s_rules"):
            await operations.delete_rule(another_internal_user_id, rule_to_delete.id)


class TestUpdateRule:
    @pytest.mark.asyncio
    async def test_update_rule(self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]) -> None:
        required_tags = list(five_tags_ids[:3])
        excluded_tags = list(five_tags_ids[3:])

        rule_to_update = await operations.create_or_update_rule(
            internal_user_id, required_tags=required_tags, excluded_tags=excluded_tags, score=13
        )

        with capture_logs() as logs:
            async with TableSizeNotChanged("s_rules"):
                updated_rule = await operations.update_rule(
                    internal_user_id,
                    rule_to_update.id,
                    required_tags=five_tags_ids[:4],
                    excluded_tags=five_tags_ids[4:],
                    score=17,
                )

        assert updated_rule.id == rule_to_update.id
        assert updated_rule.user_id == internal_user_id
        assert updated_rule.required_tags == set(five_tags_ids[:4])
        assert updated_rule.excluded_tags == set(five_tags_ids[4:])
        assert updated_rule.score == 17

        rules = await domain.get_rules(internal_user_id)

        assert rules == [updated_rule]

        assert_logs_has_business_event(
            logs,
            "rule_updated",
            user_id=internal_user_id,
            rule_id=str(rule_to_update.id),
            required_tags=list(five_tags_ids[:4]),
            excluded_tags=list(five_tags_ids[4:]),
            score=17,
        )

    @pytest.mark.asyncio
    async def test_update_not_existed_rule(
        self, internal_user_id: UserId, three_tags_ids: tuple[int, int, int]
    ) -> None:
        with capture_logs() as logs:
            async with TableSizeNotChanged("s_rules"):
                with pytest.raises(errors.NoRuleFound):
                    await operations.update_rule(
                        internal_user_id,
                        new_rule_id(),
                        required_tags=three_tags_ids[:2],
                        excluded_tags=three_tags_ids[2:],
                        score=17,
                    )

        assert_logs_has_no_business_event(logs, "rule_updated")

    @pytest.mark.asyncio
    async def test_wrong_user(
        self, internal_user_id: UserId, another_internal_user_id: UserId, three_tags_ids: tuple[int, int, int]
    ) -> None:
        rule_to_update = await operations.create_or_update_rule(
            internal_user_id, required_tags=three_tags_ids[:2], excluded_tags=[], score=13
        )

        async with TableSizeNotChanged("s_rules"):
            with pytest.raises(errors.NoRuleFound):
                await operations.update_rule(
                    another_internal_user_id,
                    rule_to_update.id,
                    required_tags=three_tags_ids[2:],
                    excluded_tags=[],
                    score=17,
                )

        rules = await domain.get_rules(internal_user_id)

        assert rules == [rule_to_update]

    @pytest.mark.asyncio
    async def test_tags_intersection(
        self, internal_user_id: UserId, five_tags_ids: tuple[int, int, int, int, int]
    ) -> None:
        rule_to_update = await operations.create_or_update_rule(
            internal_user_id, required_tags=five_tags_ids[:3], excluded_tags=[], score=13
        )

        assert set(five_tags_ids[:3]) & set(five_tags_ids[2:])

        with capture_logs() as logs:
            with pytest.raises(errors.TagsIntersection):
                await operations.update_rule(
                    internal_user_id,
                    rule_to_update.id,
                    required_tags=five_tags_ids[:3],
                    excluded_tags=five_tags_ids[2:],
                    score=17,
                )

        rules = await domain.get_rules(internal_user_id)

        assert rules == [rule_to_update]

        assert_logs_has_no_business_event(logs, "rule_updated")


# most of the logic of this function is validated in other tests
class TestGetRules:
    @pytest.mark.asyncio
    async def test_no_rules(self, internal_user_id: UserId) -> None:
        rules = await domain.get_rules(internal_user_id)

        assert rules == []


class TestCountRulesPerUser:

    @pytest.mark.asyncio
    async def test_count_rules(self, internal_user_id: UserId, another_internal_user_id: UserId) -> None:

        numbers_before = await operations.count_rules_per_user()

        await operations.create_or_update_rule(internal_user_id, required_tags=[1, 2], excluded_tags=[], score=3)
        await operations.create_or_update_rule(internal_user_id, required_tags=[], excluded_tags=[2, 3], score=5)
        await operations.create_or_update_rule(
            another_internal_user_id, required_tags=[1, 2], excluded_tags=[3, 4], score=7
        )

        numbers_after = await operations.count_rules_per_user()

        assert numbers_after[internal_user_id] == numbers_before.get(internal_user_id, 0) + 2
        assert numbers_after[another_internal_user_id] == numbers_before.get(another_internal_user_id, 0) + 1



import pytest_asyncio

from ffun.domain.entities import FeedId
from ffun.feeds.entities import Feed
from ffun.library.domain import catalog_entries, get_entry
from ffun.library.entities import Entry
from ffun.library.tests import make as l_make


@pytest_asyncio.fixture  # type: ignore
def new_entry(loaded_feed: Feed) -> Entry:
    return l_make.fake_entry(loaded_feed.source_id)


@pytest_asyncio.fixture  # type: ignore
def another_new_entry(another_loaded_feed: Feed) -> Entry:
    return l_make.fake_entry(another_loaded_feed.source_id)


@pytest_asyncio.fixture
async def cataloged_entry(loaded_feed_id: FeedId, new_entry: Entry) -> Entry:
    await catalog_entries(loaded_feed_id, [new_entry])
    return await get_entry(new_entry.id)


@pytest_asyncio.fixture
async def another_cataloged_entry(another_loaded_feed_id: FeedId, another_new_entry: Entry) -> Entry:
    await catalog_entries(another_loaded_feed_id, [another_new_entry])
    return await get_entry(another_new_entry.id)



import pytest

from ffun.core.postgresql import transaction
from ffun.core.tests.helpers import (
    TableSizeDelta,
    TableSizeNotChanged,
    assert_logs_has_business_event,
    assert_logs_has_no_business_event,
    capture_logs,
)
from ffun.domain.entities import UserId
from ffun.feeds.entities import Feed
from ffun.library.entities import Entry
from ffun.library.tests import make as l_make
from ffun.markers.entities import Marker
from ffun.markers.operations import (
    get_markers,
    remove_marker,
    remove_markers_for_entries,
    set_marker,
    tech_merge_markers,
)
from ffun.users.tests import make as u_make


class TestSetMarker:

    @pytest.mark.asyncio
    async def test_set_marker(self, internal_user_id: UserId, new_entry: Entry) -> None:
        with capture_logs() as logs:
            async with TableSizeDelta("m_markers", delta=1):
                await set_marker(internal_user_id, Marker.read, new_entry.id)

        assert_logs_has_business_event(
            logs, "marker_set", user_id=internal_user_id, entry_id=str(new_entry.id), marker=Marker.read
        )

    @pytest.mark.asyncio
    async def test_set_marker_twice(self, internal_user_id: UserId, new_entry: Entry) -> None:
        await set_marker(internal_user_id, Marker.read, new_entry.id)

        with capture_logs() as logs:
            async with TableSizeNotChanged("m_markers"):
                await set_marker(internal_user_id, Marker.read, new_entry.id)

        assert_logs_has_no_business_event(logs, "marker_set")


class TestRemoveMarker:

    @pytest.mark.asyncio
    async def test_remove_marker(self, internal_user_id: UserId, new_entry: Entry) -> None:
        await set_marker(internal_user_id, Marker.read, new_entry.id)

        with capture_logs() as logs:
            async with TableSizeDelta("m_markers", delta=-1):
                await remove_marker(internal_user_id, Marker.read, new_entry.id)

        assert_logs_has_business_event(
            logs, "marker_removed", user_id=internal_user_id, entry_id=str(new_entry.id), marker=Marker.read
        )

    @pytest.mark.asyncio
    async def test_remove_marker_twice(self, internal_user_id: UserId, new_entry: Entry) -> None:
        await set_marker(internal_user_id, Marker.read, new_entry.id)
        await remove_marker(internal_user_id, Marker.read, new_entry.id)

        with capture_logs() as logs:
            async with TableSizeNotChanged("m_markers"):
                await remove_marker(internal_user_id, Marker.read, new_entry.id)

        assert_logs_has_no_business_event(logs, "marker_removed")


class TestMergeFeeds:
    @pytest.mark.asyncio
    async def test_nothing_to_merge(self, new_entry: Entry, another_new_entry: Entry) -> None:
        async with transaction() as trx:
            await tech_merge_markers(trx, new_entry.id, another_new_entry.id)

    @pytest.mark.asyncio
    async def test_move_markers(self, loaded_feed: Feed) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed, 3)

        user_a, user_b = await u_make.n_users(2)

        await set_marker(user_a, Marker.read, entry_1.id)
        await set_marker(user_b, Marker.read, entry_2.id)

        async with TableSizeNotChanged("m_markers"):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}

    @pytest.mark.asyncio
    async def test_rewrite_links(self, loaded_feed: Feed) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed, 3)

        user_a, user_b = await u_make.n_users(2)

        await set_marker(user_a, Marker.read, entry_1.id)
        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        async with TableSizeDelta("m_markers", delta=-1):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}

    @pytest.mark.asyncio
    async def test_complex_ops(self, loaded_feed: Feed) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed, 3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await set_marker(user_a, Marker.read, entry_1.id)

        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        await set_marker(user_c, Marker.read, entry_2.id)

        async with TableSizeDelta("m_markers", delta=-1):
            async with transaction() as trx:
                await tech_merge_markers(trx, from_entry_id=entry_2.id, to_entry_id=entry_3.id)

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])
        markers_c = await get_markers(user_c, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {entry_1.id: {Marker.read}}
        assert markers_b == {entry_3.id: {Marker.read}}
        assert markers_c == {entry_3.id: {Marker.read}}


class TestRemoveMarkersForEntries:
    @pytest.mark.asyncio
    async def test_nothing_to_remove(self, loaded_feed: Feed) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed, 3)

        async with TableSizeNotChanged("m_markers"):
            await remove_markers_for_entries([entry_1.id, entry_2.id, entry_3.id])

    @pytest.mark.asyncio
    async def test_remove(self, loaded_feed: Feed) -> None:
        entry_1, entry_2, entry_3 = await l_make.n_entries_list(loaded_feed, 3)

        user_a, user_b, user_c = await u_make.n_users(3)

        await set_marker(user_a, Marker.read, entry_1.id)

        await set_marker(user_b, Marker.read, entry_2.id)
        await set_marker(user_b, Marker.read, entry_3.id)

        await set_marker(user_c, Marker.read, entry_2.id)

        async with TableSizeDelta("m_markers", delta=-3):
            await remove_markers_for_entries([entry_1.id, entry_2.id])

        markers_a = await get_markers(user_a, [entry_1.id, entry_2.id, entry_3.id])
        markers_b = await get_markers(user_b, [entry_1.id, entry_2.id, entry_3.id])
        markers_c = await get_markers(user_c, [entry_1.id, entry_2.id, entry_3.id])

        assert markers_a == {}
        assert markers_b == {entry_3.id: {Marker.read}}
        assert markers_c == {}



import uuid
from typing import Any, Iterable

from ffun.core import logging
from ffun.core.postgresql import ExecuteType, execute
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId
from ffun.feeds_collections.collections import collections
from ffun.feeds_links.entities import FeedLink

logger = logging.get_module_logger()


def row_to_feed_link(row: dict[str, Any]) -> FeedLink:
    return FeedLink(user_id=row["user_id"], feed_id=row["feed_id"], created_at=row["created_at"])


async def add_link(user_id: UserId, feed_id: FeedId) -> None:
    sql = """
        INSERT INTO fl_links (id, user_id, feed_id)
        VALUES (%(id)s, %(user_id)s, %(feed_id)s)
        ON CONFLICT (user_id, feed_id) DO NOTHING
    """

    await execute(sql, {"id": uuid.uuid4(), "user_id": user_id, "feed_id": feed_id})

    logger.business_event("feed_linked", user_id=user_id, feed_id=feed_id, in_collection=collections.has_feed(feed_id))


async def remove_link(user_id: UserId, feed_id: FeedId) -> None:
    sql = """
        DELETE FROM fl_links WHERE user_id = %(user_id)s AND feed_id = %(feed_id)s
    """

    await execute(sql, {"user_id": user_id, "feed_id": feed_id})

    logger.business_event(
        "feed_unlinked", user_id=user_id, feed_id=feed_id, in_collection=collections.has_feed(feed_id)
    )


async def tech_remove_all_links(feed_ids: Iterable[FeedId]) -> None:
    sql = """
    DELETE FROM fl_links WHERE feed_id = ANY(%(feed_ids)s)
    """

    await execute(sql, {"feed_ids": list(feed_ids)})


async def get_linked_feeds(user_id: UserId) -> list[FeedLink]:
    sql = """
        SELECT * FROM fl_links WHERE user_id = %(user_id)s
    """

    result = await execute(sql, {"user_id": user_id})

    return [row_to_feed_link(row) for row in result]


async def get_link(user_id: UserId, feed_id: FeedId) -> FeedLink | None:
    sql = """
        SELECT * FROM fl_links WHERE user_id = %(user_id)s AND feed_id = %(feed_id)s
    """

    result = await execute(sql, {"user_id": user_id, "feed_id": feed_id})

    if not result:
        return None

    return row_to_feed_link(result[0])


async def get_linked_users(feed_ids: Iterable[FeedId]) -> dict[FeedId, set[UserId]]:
    sql = """
        SELECT feed_id, user_id FROM fl_links WHERE feed_id = ANY(%(feed_ids)s)
    """

    result = await execute(sql, {"feed_ids": list(feed_ids)})

    users: dict[FeedId, set[UserId]] = {}

    for row in result:
        feed_id = row["feed_id"]
        user_id = row["user_id"]

        if feed_id not in users:
            users[feed_id] = set()

        users[feed_id].add(user_id)

    return users


async def has_linked_users(feed_id: FeedId) -> bool:
    sql = """
        SELECT 1 FROM fl_links WHERE feed_id = %(feed_id)s LIMIT 1
    """

    result = await execute(sql, {"feed_id": feed_id})

    return bool(result)


async def count_feeds_per_user() -> dict[UserId, int]:
    sql = """
        SELECT user_id, COUNT(*) FROM fl_links GROUP BY user_id
    """

    result = await execute(sql)

    return {row["user_id"]: row["count"] for row in result}


async def count_subset_feeds_per_user(feed_ids: list[FeedId]) -> dict[UserId, int]:
    sql = """
        SELECT user_id, COUNT(*) FROM fl_links WHERE feed_id = ANY(%(collection_feed_ids)s) GROUP BY user_id
    """

    result = await execute(sql, {"collection_feed_ids": feed_ids})

    return {row["user_id"]: row["count"] for row in result}


async def tech_merge_feeds(execute: ExecuteType, from_feed_id: FeedId, to_feed_id: FeedId) -> None:
    sql = """
    DELETE FROM fl_links as fll
    WHERE feed_id = %(from_feed_id)s
          AND EXISTS (SELECT 1 FROM fl_links WHERE feed_id = %(to_feed_id)s AND user_id = fll.user_id)
    """

    await execute(sql, {"from_feed_id": from_feed_id, "to_feed_id": to_feed_id})

    sql = """
    UPDATE fl_links
    SET feed_id = %(to_feed_id)s
    WHERE feed_id = %(from_feed_id)s
    """

    await execute(sql, {"from_feed_id": from_feed_id, "to_feed_id": to_feed_id})



from typing import AsyncGenerator
from unittest import mock

import fastapi
import pytest_asyncio
from httpx import ASGITransport, AsyncClient

from ffun.application import application
from ffun.core import migrations
from ffun.feeds.tests.fixtures import *  # noqa
from ffun.feeds_collections.tests.fixtures import *  # noqa
from ffun.librarian.processors.tests.fixtures import *  # noqa
from ffun.librarian.tests.fixtures import *  # noqa
from ffun.library.tests.fixtures import *  # noqa
from ffun.llms_framework.tests.fixtures import *  # noqa
from ffun.ontology.tests.fixtures import *  # noqa
from ffun.parsers.tests.fixtures import *  # noqa
from ffun.users.tests.fixtures import *  # noqa


@pytest_asyncio.fixture(scope="session", autouse=True)
async def prepare_db(
    # app: AsyncGenerator[fastapi.FastAPI, None],
) -> AsyncGenerator[None, None]:
    # database migrations may be in an inconsistent state
    await migrations.rollback_all()

    await migrations.apply_all()
    yield
    await migrations.rollback_all()


@pytest_asyncio.fixture(scope="session", autouse=True)
async def app(prepare_db: None) -> AsyncGenerator[fastapi.FastAPI, None]:
    with mock.patch("ffun.application.settings.settings.enable_api", True):
        async with application.with_app() as app:
            yield app


@pytest_asyncio.fixture()
async def client(app: fastapi.FastAPI) -> AsyncGenerator[AsyncClient, None]:
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        yield client



import json
from typing import Any

from ffun.core import logging

logger = logging.get_module_logger()


def format(data: Any) -> str:
    return json.dumps(data, ensure_ascii=False, check_circular=False, allow_nan=False, indent=None)


def parse(data: str) -> Any:
    return json.loads(data)


def finish_json(text: str, empty_value: str = '""') -> str:  # pylint: disable=too-many-branches # noqa: C901, CCR001
    stack = []

    text = text.strip()

    if text[-1] == ",":
        text = text[:-1]

    if text[-1] == ":":
        text += empty_value

    for c in text:
        if c == "{":
            stack.append(c)
            continue

        if c == "}":
            if stack[-1] == "{":
                stack.pop()
                continue

            raise NotImplementedError('For each "}" we expect "{"')

        if c == "[":
            stack.append(c)
            continue

        if c == "]":
            if stack[-1] == "[":
                stack.pop()
                continue

            raise NotImplementedError('For each "]" we expect "["')

        if c == '"':
            if stack[-1] == '"':
                stack.pop()
                continue

            stack.append(c)
            continue

    # fix

    while stack:
        c = stack.pop()

        if c == "{":
            text += "}"
            continue

        if c == "[":
            text += "]"
            continue

        if c == '"':
            text += '"'

            if stack[-1] == "{":
                text += f": {empty_value}"
                continue

            continue

    return text


def loads_with_fix(text: str) -> dict[str, Any]:
    try:
        return json.loads(finish_json(text))  # type: ignore
    except json.JSONDecodeError:
        return json.loads(finish_json(text))  # type: ignore


# TODO: remove if it will not be used anywhere
def extract_tags_from_random_json(data: Any) -> set[str]:
    if not data:
        # no tags if [], {}, ''
        return set()

    if isinstance(data, list):
        return set.union(*(extract_tags_from_random_json(item) for item in data))

    if isinstance(data, dict):
        return set.union(
            *(extract_tags_from_random_json(key) | extract_tags_from_random_json(value) for key, value in data.items())
        )

    if isinstance(data, str):
        return {data}

    return set()


def extract_tags_from_invalid_json(text: str) -> set[str]:
    logger.warning("try_to_extract_tags_from_an_invalid_ json", broken_source=text)

    # search all strings, believing that
    parts = text.split('"')

    tags: set[str] = set()

    is_tag = False

    while parts:
        value = parts.pop(0)

        if is_tag:
            tags.add(value)

        is_tag = not is_tag

    logger.info("tags_extracted", tags=tags)

    return tags



import datetime
from typing import Any

import pytest
from pytest_mock import MockerFixture

from ffun.core import json, utils
from ffun.core.tests.helpers import assert_logs, capture_logs
from ffun.domain.entities import FeedUrl, UnknownUrl
from ffun.domain.urls import normalize_classic_unknown_url, to_feed_url
from ffun.parsers.entities import EntryInfo, FeedInfo
from ffun.parsers.feed import _extract_published_at, parse_entry, parse_feed
from ffun.parsers.tests.helpers import feeds_fixtures_directory, feeds_fixtures_names


class TestParseEntry:

    def test_parse(self) -> None:
        raw_entry = {
            "title": "Entry title 1",
            "link": "/2023/07/25/news-1.html",
            "published_parsed": (2023, 7, 25, 17, 15, 0, 0, 0, -1),
            "description": "Body 1",
            "tags": [
                {"term": "tag1"},
                {"term": "tag2"},
                {"label": "tag3"},
            ],
        }

        original_url = normalize_classic_unknown_url(UnknownUrl("https://example.com/feed/"))

        assert original_url is not None

        parsed_entry = parse_entry(raw_entry, to_feed_url(original_url))

        expected_entry = EntryInfo(
            title="Entry title 1",
            body="Body 1",
            external_id="/2023/07/25/news-1.html",
            external_url=normalize_classic_unknown_url(UnknownUrl("https://example.com/2023/07/25/news-1.html")),
            published_at=datetime.datetime(2023, 7, 25, 17, 15, 0, tzinfo=datetime.timezone.utc),
            external_tags={"tag1", "tag2", "tag3"},
        )

        assert parsed_entry == expected_entry


class TestParseFeed:

    def test_error_in_feedparser(self, mocker: MockerFixture) -> None:
        url = normalize_classic_unknown_url(UnknownUrl("https://example.com/feed/"))

        assert url is not None

        mocker.patch("feedparser.parse", side_effect=ValueError("Test error in feedparser"))

        with capture_logs() as logs:
            feed_info = parse_feed("", to_feed_url(url))

        assert feed_info is None
        assert_logs(logs, error_while_parsing_feed=1)

    @pytest.mark.parametrize("raw_fixture_name", feeds_fixtures_names())
    def test_on_row_fixtures(self, raw_fixture_name: str) -> None:
        raw_fixture_path = feeds_fixtures_directory / raw_fixture_name
        expected_fixture_path = str(raw_fixture_path) + ".expected.json"

        with open(raw_fixture_path, "r", encoding="utf-8") as raw_fixture_file:
            raw_fixture = raw_fixture_file.read()

        with open(expected_fixture_path, "r", encoding="utf-8") as expected_fixture_file:
            expected_fixture = expected_fixture_file.read()

        url = normalize_classic_unknown_url(UnknownUrl("https://example.com/feed/"))

        assert url is not None

        feed_info = parse_feed(raw_fixture, to_feed_url(url))

        assert feed_info == FeedInfo.model_validate_json(expected_fixture)

    @pytest.mark.parametrize("raw_fixture_name", feeds_fixtures_names())
    def test_skip_entry_on_error(self, raw_fixture_name: str, mocker: MockerFixture) -> None:
        raw_fixture_path = feeds_fixtures_directory / raw_fixture_name
        expected_fixture_path = str(raw_fixture_path) + ".expected.json"

        with open(raw_fixture_path, "r", encoding="utf-8") as raw_fixture_file:
            raw_fixture = raw_fixture_file.read()

        with open(expected_fixture_path, "r", encoding="utf-8") as expected_fixture_file:
            expected_fixture = expected_fixture_file.read()

        url = normalize_classic_unknown_url(UnknownUrl("https://example.com/feed/"))

        assert url is not None

        call_number = {"calls": 0}

        def mocked_parse_entry(raw_entry: Any, original_url: FeedUrl) -> EntryInfo:
            call_number["calls"] += 1

            if call_number["calls"] == 1:
                raise ValueError("Test error on entry parsing")

            return parse_entry(raw_entry, original_url)

        mocker.patch("ffun.parsers.feed.parse_entry", side_effect=mocked_parse_entry)

        with capture_logs() as logs:
            feed_info = parse_feed(raw_fixture, to_feed_url(url))

        assert_logs(logs, error_while_parsing_feed_entry=1)

        # the first entry will be skipped due to the error in parsing
        parsed_expected_fixture = json.parse(expected_fixture)
        parsed_expected_fixture["entries"] = parsed_expected_fixture["entries"][1:]

        assert feed_info == FeedInfo.model_validate(parsed_expected_fixture)


class TestExtractPublishedAt:

    def test_ok(self) -> None:
        published_parsed = (2023, 7, 25, 17, 15, 0, 0, 0, -1)
        expected_published_at = datetime.datetime(2023, 7, 25, 17, 15, 0, tzinfo=datetime.timezone.utc)
        assert _extract_published_at({"published_parsed": published_parsed}) == expected_published_at

    def test_value_error(self) -> None:
        published_parsed = (1, 1, 1, 0, 0, 0, 0, 1, 0)
        assert (utils.now() - _extract_published_at({"published_parsed": published_parsed})).total_seconds() < 1



import decimal

from ffun.core import utils
from ffun.llms_framework.entities import APIKeyUsage, LLMApiKey, LLMProvider, LLMTokens, USDCost


class TestAPIKeyUsage:

    def test_spent_tokens__setuped(self) -> None:
        usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=None,
            api_key=LLMApiKey("api_key"),
            reserved_cost=USDCost(decimal.Decimal(123)),
            used_cost=USDCost(decimal.Decimal(17)),
            input_tokens=LLMTokens(100),
            output_tokens=LLMTokens(200),
            interval_started_at=utils.now(),
        )

        assert usage.cost_to_register() == 17

    def test_spent_tokens__not_setuped(self) -> None:
        usage = APIKeyUsage(
            provider=LLMProvider.test,
            user_id=None,
            api_key=LLMApiKey("api_key"),
            reserved_cost=USDCost(decimal.Decimal(132)),
            used_cost=None,
            input_tokens=LLMTokens(100),
            output_tokens=LLMTokens(200),
            interval_started_at=utils.now(),
        )

        assert usage.cost_to_register() == 132



import pytest

from ffun.core.entities import BaseEntity
from ffun.librarian.text_cleaners import clear_html, clear_nothing


class TestClearNothing:

    @pytest.mark.parametrize(
        "input_text, expected_text",
        [
            ("", ""),
            ("some-text some-text", "some-text some-text"),
            ("some <tagged> text</tagged>", "some <tagged> text</tagged>"),
        ],
    )
    def test(self, input_text: str, expected_text: str) -> None:
        assert clear_nothing(input_text) == expected_text


class TestClearHTML:

    def test(self) -> None:
        input_text = """
<h1>Some title</h1>
<div>
<script>some script</script>
<style>some style</style>
        Text to keep
<meta content="some meta"/>
<iframe>some iframe</iframe>
<img alt="some img" src="some src" title="some title" width="100" height="100" class="some class" style="some style" />
</div>

<h2>Some other title 2</h2>
        <h3>Some other title 3</h3>
        <h4>Some other title 4</h4>
        <h5>Some other title 5</h5>
        <h6>Some other title 6</h6>

<a href="https://example.com">Some link</a>

Just text

<p>Some paragraph</p>

<ul>
<li>Some list item 1</li>
<li>Some list item 2</li>
<li><ol><li>Some nestedlist item 3</li>
<li>Some nestedlist item 4</li></ol></li>
</ul>
"""
        expected_text = """<h1>Some title</h1> Text to keep <h2>Some other title 2</h2> <h3>Some other title 3</h3> <h4>Some other title 4</h4> <h5>Some other title 5</h5> <h6>Some other title 6</h6> <a href="https://example.com">Some link</a> Just text <p>Some paragraph</p> <ul> <li>Some list item 1</li> <li>Some list item 2</li> <li><ol><li>Some nestedlist item 3</li> <li>Some nestedlist item 4</li></ol></li> </ul>"""  # noqa

        assert clear_html(input_text) == expected_text

    def test_broken_unicode(self) -> None:
        input_text = "some text &#55358;bla-bla"

        expected_text = "some text bla-bla"

        processed_text = clear_html(input_text)

        assert processed_text == expected_text

        # parser resulting string with pydantic
        class TestModel(BaseEntity):
            text: str

        model = TestModel(text=processed_text)

        assert model.text == expected_text



import enum
from typing import Any, Protocol, Union

import pydantic

from ffun.core.entities import BaseEntity
from ffun.domain.entities import AbsoluteUrl, FeedUrl, UnknownUrl
from ffun.parsers import entities as p_entities


class Discoverer(Protocol):
    async def __call__(self, context: "Context") -> tuple["Context", Union[None, "Result"]]:
        pass


class Status(enum.StrEnum):
    feeds_found = "feeds_found"
    incorrect_url = "incorrect_url"
    cannot_access_url = "cannot_access_url"
    not_html = "not_html"
    no_feeds_found = "no_feeds_found"


class Context(BaseEntity):
    raw_url: UnknownUrl
    url: FeedUrl | None = None
    content: str | None = None
    soup: Any | None = None
    depth: int = 1
    candidate_urls: set[AbsoluteUrl] = pydantic.Field(default_factory=set)
    discoverers: list[Any] = pydantic.Field(default_factory=list)

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)


class Result(BaseEntity):
    feeds: list[p_entities.FeedInfo]
    status: Status



from typing import TYPE_CHECKING, Any

from sentry_sdk import init as initialize_sentry

if TYPE_CHECKING:
    from sentry_sdk._types import Event

from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration

import ffun
from ffun.core.errors import Error


def improve_fingerprint(event: "Event", hint: dict[str, Any]) -> "Event":
    if "exc_info" not in hint:
        return event

    _, e, _ = hint["exc_info"]

    if isinstance(e, Error) and e.fingerprint is not None:
        event["fingerprint"] = ["{{ default }}", e.fingerprint]

    return event


def before_send(event: "Event", hint: dict[str, Any]) -> "Event":
    return improve_fingerprint(event, hint)


def initialize(dsn: str, sample_rate: float, environment: str) -> None:
    initialize_sentry(
        dsn=dsn,
        sample_rate=sample_rate,
        # Background worker traces are not useful
        # because most of the reported ones are related to previous tasks.
        # It can be fixed by reinitializing Sentry's transaction context
        # but it is not worth the effort for now.
        traces_sample_rate=None,
        # Disable breadcrumbs by the same cause — does not work well with background workers.
        max_breadcrumbs=0,
        # Without this config Sentry miss important frames from stacktraces
        add_full_stack=True,
        attach_stacktrace=True,
        before_send=before_send,
        environment=environment,
        # set the correct project root directory
        project_root=ffun.__path__[0],
        # disable ALL automatically enabled integrations, because they periodically cause issues
        # like false positive errors for correctly handled exceptions in OpenAI integration
        auto_enabling_integrations=False,
        integrations=[
            StarletteIntegration(transaction_style="endpoint"),
            FastApiIntegration(transaction_style="endpoint"),
        ],
        disabled_integrations=[
            # disable default logging integration to use specialized structlog-sentry processor
            LoggingIntegration(event_level=None, level=None),
        ],
    )



import contextlib
import functools
from typing import Generator, Sequence

import openai
import tiktoken
from openai.types.chat import ChatCompletionMessageParam

from ffun.core import logging
from ffun.llms_framework import domain as llmsf_domain
from ffun.llms_framework import errors as llmsf_errors
from ffun.llms_framework.entities import KeyStatus, LLMConfiguration, LLMProvider
from ffun.llms_framework.keys_statuses import Statuses
from ffun.llms_framework.provider_interface import ChatRequest, ChatResponse, ProviderInterface
from ffun.openai.settings import settings

logger = logging.get_module_logger()


class OpenAIChatRequest(ChatRequest):
    messages: list[ChatCompletionMessageParam]


def _client(api_key: str) -> openai.AsyncOpenAI:
    return openai.AsyncOpenAI(api_key=api_key, base_url=settings.api_entry_point, timeout=settings.api_timeout)


class OpenAIChatResponse(ChatResponse):
    content: str
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

    def response_content(self) -> str:
        return self.content

    def input_tokens(self) -> int:
        return self.prompt_tokens

    def output_tokens(self) -> int:
        return self.completion_tokens


@functools.cache
def _get_encoding(model: str) -> tiktoken.Encoding:
    # TODO: rollback after tiktoken add support for new models
    #       https://github.com/openai/tiktoken/issues/395
    #       https://github.com/openai/tiktoken/pull/396
    if "gpt-4.1" in model:
        return tiktoken.encoding_for_model("gpt-4o")

    return tiktoken.encoding_for_model(model)


@contextlib.contextmanager
def track_key_status(key: str, statuses: Statuses) -> Generator[None, None, None]:
    try:
        yield
        statuses.set(key, KeyStatus.works)
    except openai.AuthenticationError:
        statuses.set(key, KeyStatus.broken)
        raise
    except openai.RateLimitError:
        statuses.set(key, KeyStatus.quota)
        raise
    except openai.PermissionDeniedError:
        statuses.set(key, KeyStatus.broken)
        raise
    except openai.APIError:
        statuses.set(key, KeyStatus.unknown)
        raise


# TODO: tests
class OpenAIInterface(ProviderInterface):
    provider = LLMProvider.openai

    additional_tokens_per_message: int = 10

    def estimate_tokens(self, config: LLMConfiguration, text: str) -> int:
        encoding = _get_encoding(config.model)

        system_tokens = (
            self.additional_tokens_per_message + len(encoding.encode("system")) + len(encoding.encode(config.system))
        )

        text_tokens = self.additional_tokens_per_message + len(encoding.encode("user")) + len(encoding.encode(text))

        return system_tokens + text_tokens

    async def chat_request(  # type: ignore
        self, config: LLMConfiguration, api_key: str, request: OpenAIChatRequest
    ) -> OpenAIChatResponse:
        try:
            with track_key_status(api_key, self.api_keys_statuses):
                answer = await _client(api_key=api_key).chat.completions.create(
                    model=config.model,
                    temperature=float(config.temperature),
                    max_tokens=config.max_return_tokens,
                    top_p=float(config.top_p),
                    presence_penalty=float(config.presence_penalty),
                    frequency_penalty=float(config.frequency_penalty),
                    messages=request.messages,
                )
        except openai.APIError as e:
            message = str(e)
            logger.info("openai_api_error", message=message)
            raise llmsf_errors.TemporaryError(message=message) from e

        logger.info("openai_response")

        assert answer.choices[0].message.content is not None
        assert answer.usage is not None

        content = answer.choices[0].message.content

        return OpenAIChatResponse(
            content=content,
            prompt_tokens=answer.usage.prompt_tokens,
            completion_tokens=answer.usage.completion_tokens,
            total_tokens=answer.usage.total_tokens,
        )

    def prepare_requests(self, config: LLMConfiguration, text: str) -> Sequence[OpenAIChatRequest]:

        parts = llmsf_domain.split_text_according_to_tokens(llm=self, llm_config=config, text=text)

        requests = []

        for part in parts:
            request = OpenAIChatRequest(
                messages=[{"role": "system", "content": config.system}, {"role": "user", "content": part}]
            )

            requests.append(request)

        return requests

    async def check_api_key(self, config: LLMConfiguration, api_key: str) -> KeyStatus:
        with track_key_status(api_key, self.api_keys_statuses):
            try:
                await _client(api_key=api_key).models.list()
            except openai.APIError:
                pass

        return self.api_keys_statuses.get(api_key)


provider = OpenAIInterface()



import sys

import tabulate

from ffun.core import utils
from ffun.librarian.background_processors import processors as ln_processors
from ffun.ontology import domain as o_domain
from ffun.processors_quality.entities import ProcessorResult, ProcessorResultDiff
from ffun.processors_quality.knowlege_base import KnowlegeBase, id_to_name

processors = {info.processor.name: info.processor for info in ln_processors}


async def run_processor(kb: KnowlegeBase, processor_name: str, entry_id: int) -> ProcessorResult:

    entry = kb.get_news_entry(entry_id)

    processor = processors[processor_name]

    raw_tags = await processor.process(entry)

    raw_tags_to_uids = await o_domain.normalize_tags(raw_tags)

    tags = set(raw_tags_to_uids.values())

    result = ProcessorResult(
        tags=list(sorted(tags)),
        created_at=utils.now(),
    )

    return result


def diff_processor_results(kb: KnowlegeBase, processor_name: str, entry_ids: list[int]) -> list[ProcessorResultDiff]:

    diffs: list[ProcessorResultDiff] = []

    for entry_id in sorted(entry_ids):
        expected = kb.get_expected_tags(processor_name, entry_id)
        actual = kb.get_actual_results(processor_name, entry_id)
        last = kb.get_last_results(processor_name, entry_id)

        actual_tags = set(actual.tags)
        last_tags = set(last.tags)

        diff = ProcessorResultDiff(
            entry_id=entry_id,
            must_have_total=len(expected.must_have),
            should_have_total=len(expected.should_have),
            actual_total=len(actual_tags),
            actual_must_have_found=len(expected.must_have & actual_tags),
            actual_must_have_missing=list(expected.must_have - actual_tags),
            actual_should_have_found=len(expected.should_have & actual_tags),
            actual_has_and_last_not=list(actual_tags - last_tags),
            last_total=len(last_tags),
            last_must_have_found=len(expected.must_have & last_tags),
            last_must_have_missing=list(expected.must_have - last_tags),
            last_should_have_found=len(expected.should_have & last_tags),
            last_has_and_actual_not=list(last_tags - actual_tags),
        )

        diffs.append(diff)

    return diffs


def display_diffs(diffs: list[ProcessorResultDiff], show_tag_diffs: bool) -> None:  # noqa: CCR001
    table = []

    if not show_tag_diffs:
        headers = ["entry id", "total", "must have", "should have", "should have %"]
    else:
        headers = [
            "entry id",
            "total",
            "must have",
            "should have",
            "should have %",
            "actual has and last not",
            "last has and actual not",
        ]

    missing_must_have = 0

    should_diffs = []
    total_diffs = []

    for diff in diffs:

        if diff.actual_must_have_found != diff.must_have_total:
            raise NotImplementedError(
                f'Currently we expect that actual will always have "must" tags, tags: {diff.actual_must_have_missing}'
            )  # noqa: E501

        if diff.actual_must_have_found == diff.last_must_have_found:
            must_have = "ok"
        elif diff.actual_must_have_found > diff.last_must_have_found:
            must_have = ", ".join(f"?{tag}" for tag in diff.last_must_have_missing)
            missing_must_have += 1
        else:
            raise NotImplementedError("We should not reach this point")

        should_delta = diff.last_should_have_found - diff.actual_should_have_found

        should_have = f"{should_delta:+} / {diff.should_have_total}"

        if diff.should_have_total == 0:
            should_diff = 0.0
        else:
            should_diff = should_delta / diff.should_have_total

        should_diffs.append(should_diff)

        should_have_percent = f"{should_diff:+.2%}"

        if diff.actual_total == 0:
            total_fraction = 1.0
        else:
            total_fraction = diff.last_total / diff.actual_total - 1

        total = f"{diff.actual_total} vs {diff.last_total} [{total_fraction:+.2%}]"

        total_diffs.append(total_fraction)

        row = [id_to_name(diff.entry_id), total, must_have, should_have, should_have_percent]

        if show_tag_diffs:
            actual_has_and_last_not = "\n".join(diff.actual_has_and_last_not)
            last_has_and_actual_not = "\n".join(diff.last_has_and_actual_not)
            row.extend([actual_has_and_last_not, last_has_and_actual_not])

        table.append(row)

    sys.stdout.write("\n")
    sys.stdout.write(tabulate.tabulate(table, headers=headers, tablefmt="grid"))
    sys.stdout.write("\n")
    sys.stdout.write("\n")

    if missing_must_have:
        sys.stdout.write(f"ERROR: entries with m must have tags: {missing_must_have}\n\n")

    # should diffs
    should_diffs = list(sorted(should_diffs))

    sys.stdout.write("should diffs:\n")

    should_median = should_diffs[len(should_diffs) // 2]

    sys.stdout.write(f"worst: {should_diffs[0]:.2%}, median: {should_median:.2%}, best: {should_diffs[-1]:.2%}\n")

    should_diff_average = sum(should_diffs) / len(should_diffs)

    sys.stdout.write(f"average: {should_diff_average:.2%}\n\n")

    # total diffs
    total_diffs = list(sorted(total_diffs))

    sys.stdout.write("total diffs:\n")

    sys.stdout.write(
        f"worst: {total_diffs[0]:.2%}, median: {total_diffs[len(total_diffs) // 2]:.2%}, best: {total_diffs[-1]:.2%}\n"
    )

    total_diff_average = sum(total_diffs) / len(total_diffs)

    sys.stdout.write(f"average: {total_diff_average:.2%}\n\n")



from typing import Type
from unittest.mock import MagicMock

import openai
import pytest

from ffun.llms_framework.entities import KeyStatus
from ffun.llms_framework.keys_statuses import Statuses
from ffun.openai.provider_interface import track_key_status


class TestTrackKeyStatus:
    def test_works(self, api_key_statuses: Statuses) -> None:
        with track_key_status("key_1", api_key_statuses):
            pass

        assert api_key_statuses.get("key_1") == KeyStatus.works

    @pytest.mark.parametrize("exception", [openai.AuthenticationError, openai.PermissionDeniedError])
    def test_authentication_error(self, exception: Type[Exception], api_key_statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", api_key_statuses):
                raise exception(message="test-message", response=MagicMock(), body=MagicMock())  # type: ignore

        assert api_key_statuses.get("key_1") == KeyStatus.broken

    @pytest.mark.parametrize("exception", [openai.RateLimitError])
    def test_quota_error(self, exception: Type[Exception], api_key_statuses: Statuses) -> None:
        with pytest.raises(exception):
            with track_key_status("key_1", api_key_statuses):
                raise exception(message="test-message", response=MagicMock(), body=MagicMock())  # type: ignore

        assert api_key_statuses.get("key_1") == KeyStatus.quota



from urllib.parse import urlparse

from ffun.core import logging
from ffun.librarian.processors import base
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag, TagCategory

logger = logging.get_module_logger()


def domain_to_parts(domain: str) -> list[str]:
    if domain.startswith("www."):
        domain = domain[4:]

    parts = []

    while domain:
        if "." not in domain:
            # do not add top level domains like .com
            # they are always on top of the tags list
            # and it is very rare situation when user wants to score by them
            break

        parts.append(domain)

        domain = domain.split(".", 1)[1]

    return parts


class Processor(base.Processor):
    __slots__ = ()

    async def process(self, entry: Entry) -> list[ProcessorTag]:
        tags: list[ProcessorTag] = []

        if not entry.external_url:
            return tags

        try:
            parsed_url = urlparse(entry.external_url)
        except Exception:
            # TODO: log error somewhere, link it to the entry
            logger.exception("failed_to_parse_url", url=entry.external_url)
            return tags

        domain = parsed_url.netloc

        for subdomain in domain_to_parts(domain):
            tags.append(
                ProcessorTag(
                    raw_uid=subdomain,
                    link=parsed_url.scheme + "://" + subdomain,
                    categories={TagCategory.network_domain},
                )
            )

        return tags



"""
scores-tables
"""

from typing import Any

from psycopg import Connection
from yoyo import step

__depends__: set[str] = set()

steps = [step("")]


sql_create_rules_table = """
CREATE TABLE s_rules (
    id UUID PRIMARY KEY,
    user_id UUID NOT NULL,
    tags BIGINT[] NOT NULL,
    key TEXT NOT NULL,
    score INTEGER NOT NULL,
    created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT NOW()
)
"""

sql_users_key_index = """
CREATE UNIQUE INDEX idx_s_rules_user_id_key ON s_rules (user_id, key);
"""


def apply_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute(sql_create_rules_table)
    cursor.execute(sql_users_key_index)


def rollback_step(conn: Connection[dict[str, Any]]) -> None:
    cursor = conn.cursor()
    cursor.execute("DROP TABLE s_rules")


steps = [step(apply_step, rollback_step)]



import contextlib
from typing import Any, Generator

import httpx

from ffun.core import logging
from ffun.google import errors
from ffun.google.entities import GenerationConfig, GoogleChatRequest, GoogleChatResponse
from ffun.google.settings import settings

logger = logging.get_module_logger()

_safety_categories = [
    "HARM_CATEGORY_HATE_SPEECH",
    "HARM_CATEGORY_SEXUALLY_EXPLICIT",
    "HARM_CATEGORY_DANGEROUS_CONTENT",
    "HARM_CATEGORY_HARASSMENT",
]


class Client:
    __slots__ = ("_api_key", "entry_point", "model")

    def __init__(self, api_key: str, entry_point: str = settings.gemini_api_entry_point) -> None:
        self._api_key = api_key
        self.entry_point = entry_point

    @contextlib.contextmanager
    def _handle_network_errors(self) -> Generator[None, None, None]:
        try:
            yield
        except Exception as e:
            raise errors.UnknownError(message=str(e)) from e

    def _handle_response_status_errors(self, response: httpx.Response) -> None:
        if response.status_code == 429:
            raise errors.QuotaError(message=response.content, status_code=response.status_code)

        if response.status_code in (401, 403):
            raise errors.AuthError(message=response.content, status_code=response.status_code)

        if response.status_code != 200:
            raise errors.UnknownError(message=response.content, status_code=response.status_code)

    # See https://ai.google.dev/api/generate-content#FinishReason
    def _handle_finish_reason(self, finish_reason: str) -> None:
        match finish_reason:
            case "FINISH_REASON_UNSPECIFIED" | "STOP" | "MAX_TOKENS" | "OTHER":
                pass
            case "SAFETY" | "RECITATION" | "LANGUAGE" | "BLOCKLIST" | "SPII" | "IMAGE_SAFETY":
                # we may want to stop processing the response here
                # let's wait it there will be any issues with just passing it
                pass
            case "PROHIBITED_CONTENT":
                raise errors.ProhibitedContentError()
            case "MALFORMED_FUNCTION_CALL":
                raise errors.MalformedFunctionCallError()
            case reason:
                raise errors.UnknownFinishReasonError(reason=reason)

    # See https://ai.google.dev/api/generate-content#PromptFeedback
    def _handle_prompt_feedback(self, prompt_feedback: dict[str, Any] | None) -> None:
        if prompt_feedback is None:
            return

        if "blockReason" not in prompt_feedback:
            return

        raise errors.PromptBlocked(reason=prompt_feedback["blockReason"])

    async def generate_content(
        self,
        model: str,
        config: GenerationConfig,
        request: GoogleChatRequest,
        timeout: float = settings.gemini_api_timeout,
    ) -> GoogleChatResponse:

        headers = {"Content-Type": "application/json"}

        url = f"{self.entry_point}/models/{model}:generateContent?key={self._api_key}"

        http_request = {
            "systemInstruction": request.system.to_api(),
            "contents": [message.to_api() for message in request.messages],
            "generationConfig": config.to_api(),
            "safetySettings": [{"category": category, "threshold": "BLOCK_NONE"} for category in _safety_categories],
        }

        with self._handle_network_errors():
            async with httpx.AsyncClient(headers=headers, timeout=timeout) as client:
                response = await client.post(url, json=http_request)

        self._handle_response_status_errors(response)

        response_data = response.json()

        # There is cases where no candidates are returned
        # => Check prompt feedback before processing candidates
        # See https://ai.google.dev/api/generate-content#generatecontentresponse
        # Also, it looks like `promptFeedback` is optional => we use `get` instead of `[]`
        self._handle_prompt_feedback(response_data.get("promptFeedback"))

        candidate = response_data["candidates"][0]

        self._handle_finish_reason(candidate["finishReason"])

        return GoogleChatResponse(
            content=candidate["content"]["parts"][0]["text"],
            prompt_tokens=response_data["usageMetadata"]["promptTokenCount"],
            completion_tokens=response_data["usageMetadata"]["candidatesTokenCount"],
            total_tokens=response_data["usageMetadata"]["totalTokenCount"],
        )

    async def list_models(self, timeout: float = settings.gemini_api_timeout) -> list[dict[str, Any]]:
        headers = {"Content-Type": "application/json"}

        url = f"{self.entry_point}/models?key={self._api_key}"

        with self._handle_network_errors():
            async with httpx.AsyncClient(headers=headers, timeout=timeout) as client:
                response = await client.get(url)

        self._handle_response_status_errors(response)

        # TODO: introduce entity class for models
        return response.json()["models"]  # type: ignore



import datetime
from typing import Any

import pydantic

from ffun.domain.entities import AbsoluteUrl, FeedUrl, UrlUid


class EntryInfo(pydantic.BaseModel):
    title: str
    body: str
    external_id: str
    external_url: AbsoluteUrl | None
    external_tags: set[str]
    published_at: datetime.datetime

    def log_info(self) -> dict[str, Any]:
        return {"title": self.title, "external_url": self.external_url}


class FeedInfo(pydantic.BaseModel):
    url: FeedUrl
    title: str
    description: str
    uid: UrlUid

    entries: list[EntryInfo]

    def log_info(self) -> dict[str, Any]:
        return {"title": self.title, "url": self.url}



import asyncio
import datetime
import random
import uuid

import pytest
import pytest_asyncio

from ffun.core import utils
from ffun.core.tests.helpers import Delta, TableSizeDelta, TableSizeNotChanged
from ffun.domain.domain import new_feed_id
from ffun.domain.entities import FeedId
from ffun.domain.urls import str_to_absolute_url, url_to_uid
from ffun.feeds import errors
from ffun.feeds.domain import get_feed, save_feeds
from ffun.feeds.entities import Feed, FeedError, FeedState
from ffun.feeds.operations import (
    count_total_feeds,
    count_total_feeds_per_last_error,
    count_total_feeds_per_state,
    get_feed_ids_by_uids,
    get_feeds,
    get_next_feeds_to_load,
    get_orphaned_feeds,
    get_source_ids,
    mark_feed_as_failed,
    mark_feed_as_loaded,
    mark_feed_as_orphaned,
    save_feed,
    tech_remove_feed,
    update_feed_info,
)
from ffun.feeds.tests import make


class TestSaveFeed:
    @pytest.mark.asyncio
    async def test_new_feed(self, new_feed: Feed) -> None:
        feed_id = await save_feed(new_feed)
        assert feed_id == new_feed.id
        assert await get_feed(new_feed.id) == new_feed

    @pytest.mark.asyncio
    async def test_existed_feed(self, new_feed: Feed) -> None:
        original_feed_id = await save_feed(new_feed)

        cloned_feed = new_feed.replace(id=uuid.uuid4())

        saved_feed_id = await save_feed(cloned_feed)

        assert original_feed_id == saved_feed_id

        assert await get_feed(original_feed_id) == new_feed

        with pytest.raises(errors.NoFeedFound):
            await get_feed(cloned_feed.id)

    @pytest.mark.asyncio
    async def test_same_uid_different_urls(self, new_feed: Feed) -> None:
        url_part = uuid.uuid4().hex

        feed_1 = new_feed.replace(url=f"http://example.com/{url_part}")
        feed_2 = new_feed.replace(id=uuid.uuid4(), url=f"https://example.com/{url_part}")

        feed_1_id = await save_feed(feed_1)
        feed_2_id = await save_feed(feed_2)

        assert feed_1_id == feed_2_id

        assert await get_feed(feed_1_id) == feed_1

        with pytest.raises(errors.NoFeedFound):
            await get_feed(feed_2.id)


class TestGetNextFeedToLoad:
    @pytest.mark.asyncio
    async def test_find_new_feed(self, saved_feed_id: FeedId) -> None:
        now = utils.now()

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        feeds = {feed.id: feed for feed in found_feeds}

        found_feed = feeds[saved_feed_id]

        assert now < found_feed.load_attempted_at

        loaded_feed = await get_feed(saved_feed_id)

        assert loaded_feed.load_attempted_at == found_feed.load_attempted_at

    @pytest.mark.asyncio
    async def test_skip_choosen_feeds(self, saved_feed_id: FeedId) -> None:
        now = utils.now()

        await get_next_feeds_to_load(number=100500, loaded_before=now)

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        assert not found_feeds

    @pytest.mark.asyncio
    async def test_find_choosen_after_time_passes(self, saved_feed_id: FeedId) -> None:
        await get_next_feeds_to_load(number=100500, loaded_before=utils.now())

        now = utils.now()

        found_feeds = await get_next_feeds_to_load(number=100500, loaded_before=now)

        feeds = {feed.id: feed for feed in found_feeds}

        found_feed = feeds[saved_feed_id]

        assert now < found_feed.load_attempted_at

        loaded_feed = await get_feed(saved_feed_id)

        assert loaded_feed.load_attempted_at == found_feed.load_attempted_at

    @pytest.mark.asyncio
    async def test_limit(self) -> None:
        now = utils.now()

        # mark existed feeds as chosen
        await get_next_feeds_to_load(number=100500, loaded_before=now)

        n = 10
        m = 3

        feed_ids = set(await save_feeds([await make.fake_feed() for _ in range(n)]))

        while feed_ids:
            found_feeds = await get_next_feeds_to_load(number=m, loaded_before=now)

            assert len(found_feeds) == min(m, len(feed_ids))

            found_feed_ids = {feed.id for feed in found_feeds}

            assert found_feed_ids <= feed_ids

            feed_ids -= found_feed_ids

    @pytest.mark.asyncio
    async def test_concurent_run(self) -> None:
        now = utils.now()

        # mark existed feeds as chosen
        await get_next_feeds_to_load(number=100500, loaded_before=now)

        n = 10
        m = 3

        feed_ids = set(await save_feeds([await make.fake_feed() for _ in range(n)]))

        tasks = [get_next_feeds_to_load(number=m, loaded_before=now) for _ in range(n // m + 1)]

        results = await asyncio.gather(*tasks)

        found_feed_ids = {feed.id for feeds in results for feed in feeds}

        assert found_feed_ids == feed_ids


class TestUpdateFeedInfo:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        new_title = make.fake_title()
        new_description = make.fake_description()

        await update_feed_info(feed_id=saved_feed.id, title=new_title, description=new_description)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.title == new_title
        assert updated_feed.description == new_description


class TestMarkFeedAsLoaded:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        now = utils.now()

        await mark_feed_as_loaded(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.loaded_at is not None
        assert now < updated_feed.loaded_at
        assert updated_feed.last_error is None
        assert updated_feed.state == FeedState.loaded

    @pytest.mark.asyncio
    async def test_set_loaded_at(self, saved_feed: Feed) -> None:
        loaded_at = utils.now() - datetime.timedelta(days=1)

        await mark_feed_as_loaded(feed_id=saved_feed.id, loaded_at=loaded_at)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.loaded_at is not None
        assert loaded_at == updated_feed.loaded_at
        assert updated_feed.last_error is None
        assert updated_feed.state == FeedState.loaded

    @pytest.mark.asyncio
    async def test_reset_error_state(self, saved_feed: Feed) -> None:
        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=random.choice(list(FeedError)))

        await mark_feed_as_loaded(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.last_error is None


class TestMarkFeedAsFailed:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        error = random.choice(list(FeedError))

        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=error)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.state == FeedState.damaged
        assert updated_feed.last_error == error


class TestMarkFeedAsOrphaned:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        await mark_feed_as_orphaned(feed_id=saved_feed.id)

        assert saved_feed.loaded_at is None

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.loaded_at is None
        assert updated_feed.last_error is None
        assert updated_feed.state == FeedState.orphaned

    @pytest.mark.asyncio
    async def test_reset_error_state(self, saved_feed: Feed) -> None:
        await mark_feed_as_failed(feed_id=saved_feed.id, state=FeedState.damaged, error=random.choice(list(FeedError)))

        await mark_feed_as_orphaned(feed_id=saved_feed.id)

        updated_feed = await get_feed(saved_feed.id)

        assert updated_feed.last_error is None


class TestGetOrphanedFeeds:

    @pytest_asyncio.fixture(autouse=True)
    async def cleanup_orphaned_feeds(self) -> None:
        orphanes = await get_orphaned_feeds(limit=10000, loaded_before=utils.now())

        for orphane_id in orphanes:
            await tech_remove_feed(orphane_id)

    @pytest.mark.asyncio
    async def test_no_orphans(self) -> None:
        await make.n_feeds(3)

        orphaned_feeds = await get_orphaned_feeds(limit=10, loaded_before=utils.now())

        assert not orphaned_feeds

    @pytest.mark.asyncio
    async def test_orphans(self) -> None:

        loaded_before = utils.now() - datetime.timedelta(days=2)
        check_at = loaded_before + datetime.timedelta(days=1)

        feeds = await make.n_feeds(5)

        await mark_feed_as_orphaned(feed_id=feeds[0].id)
        assert feeds[0].loaded_at is None

        await mark_feed_as_loaded(feed_id=feeds[1].id)
        await mark_feed_as_orphaned(feed_id=feeds[1].id)

        await mark_feed_as_loaded(feed_id=feeds[2].id, loaded_at=loaded_before)
        await mark_feed_as_orphaned(feed_id=feeds[2].id)

        # test all found
        orphanes = await get_orphaned_feeds(limit=10, loaded_before=utils.now())
        assert set(orphanes) == {feeds[0].id, feeds[1].id, feeds[2].id}

        # test limit
        orphanes = await get_orphaned_feeds(limit=2, loaded_before=utils.now())
        assert len(orphanes) == 2
        assert set(orphanes) <= {feeds[0].id, feeds[1].id, feeds[2].id}

        # test loaded_before
        orphanes = await get_orphaned_feeds(limit=10, loaded_before=check_at)
        assert set(orphanes) == {feeds[0].id, feeds[2].id}


class TestGetFeeds:
    @pytest.mark.asyncio
    async def test(self) -> None:
        n = 3

        feed_ids = []

        for _ in range(n + 2):
            raw_feed = await make.fake_feed()
            feed_id = await save_feed(raw_feed)
            feed_ids.append(feed_id)

        loaded_feeds = await get_feeds(feed_ids[1:-1])

        assert len(loaded_feeds) == n

        assert set(feed_ids[1:-1]) == {feed.id for feed in loaded_feeds}


class TestGetFeedIdsByUids:

    @pytest.mark.asyncio
    async def test_no_uids_passed(self) -> None:
        ids = await get_feed_ids_by_uids([])

        assert ids == {}

    @pytest.mark.asyncio
    async def test_load(self, saved_feed: Feed, another_saved_feed: Feed) -> None:
        uids = [
            url_to_uid(saved_feed.url),
            url_to_uid(another_saved_feed.url),
            url_to_uid(str_to_absolute_url(f"http://{uuid.uuid4().hex}.com/")),
        ]

        ids = await get_feed_ids_by_uids(uids)

        assert len(ids) == 2

        assert ids[uids[0]] == saved_feed.id
        assert ids[uids[1]] == another_saved_feed.id


class TestGetSourceIds:

    @pytest.mark.asyncio
    async def test_no_uids_passed(self) -> None:
        async with TableSizeNotChanged("f_sources"):
            ids = await get_source_ids([])

        assert ids == {}

    @pytest.mark.asyncio
    async def test_all_new(self) -> None:
        n = 3

        uids = [f"{uuid.uuid4().hex}.com" for _ in range(n)]

        async with TableSizeDelta("f_sources", delta=n):
            ids = await get_source_ids(uids)

        assert set(ids.keys()) == set(uids)

    @pytest.mark.asyncio
    async def test_partially_saved(self) -> None:
        n = 5
        m = 3

        uids = [f"{uuid.uuid4().hex}.com" for _ in range(n)]

        await get_source_ids(uids[:m])

        async with TableSizeDelta("f_sources", delta=n - m):
            ids = await get_source_ids(uids)

        assert set(ids.keys()) == set(uids)


class TestTechRemoveFeed:
    @pytest.mark.asyncio
    async def test(self, saved_feed: Feed) -> None:
        async with TableSizeDelta("f_feeds", delta=-1):
            await tech_remove_feed(saved_feed.id)

        with pytest.raises(errors.NoFeedFound):
            await get_feed(saved_feed.id)

    @pytest.mark.asyncio
    async def test_no_feed(self) -> None:
        await tech_remove_feed(new_feed_id())


class TestCountTotalFeeds:

    @pytest.mark.asyncio
    async def test(self) -> None:
        async with Delta(count_total_feeds, delta=3):
            await save_feeds([await make.fake_feed() for _ in range(3)])


class TestCountTotalFeedsPerState:

    @pytest.mark.asyncio
    async def test(self) -> None:
        numbers_before = await count_total_feeds_per_state()

        await save_feeds(
            [
                await make.fake_feed(state=FeedState.loaded),
                await make.fake_feed(state=FeedState.damaged),
                await make.fake_feed(state=FeedState.loaded),
                await make.fake_feed(state=FeedState.orphaned),
                await make.fake_feed(state=FeedState.loaded),
            ]
        )

        numbers_after = await count_total_feeds_per_state()

        assert numbers_after[FeedState.loaded] == numbers_before.get(FeedState.loaded, 0) + 3
        assert numbers_after[FeedState.damaged] == numbers_before.get(FeedState.damaged, 0) + 1
        assert numbers_after[FeedState.orphaned] == numbers_before.get(FeedState.orphaned, 0) + 1


class TestCountTotalFeedsPerLastError:

    @pytest.mark.asyncio
    async def test(self) -> None:
        numbers_before = await count_total_feeds_per_last_error()

        feeds = await make.n_feeds(5)

        await mark_feed_as_loaded(feed_id=feeds[0].id)
        await mark_feed_as_failed(
            feed_id=feeds[1].id, state=FeedState.damaged, error=FeedError.network_wrong_ssl_version
        )
        await mark_feed_as_loaded(feed_id=feeds[2].id)
        await mark_feed_as_failed(
            feed_id=feeds[3].id, state=FeedState.damaged, error=FeedError.network_wrong_ssl_version
        )
        await mark_feed_as_failed(
            feed_id=feeds[4].id, state=FeedState.damaged, error=FeedError.protocol_no_entries_in_feed
        )

        numbers_after = await count_total_feeds_per_last_error()

        assert (
            numbers_after[FeedError.network_wrong_ssl_version]
            == numbers_before[FeedError.network_wrong_ssl_version] + 2
        )
        assert (
            numbers_after[FeedError.protocol_no_entries_in_feed]
            == numbers_before[FeedError.protocol_no_entries_in_feed] + 1
        )



import contextlib
import datetime
from decimal import Decimal
from typing import Any, AsyncGenerator, Collection, Iterable, Protocol

from ffun.core import logging
from ffun.domain.entities import UserId
from ffun.feeds.entities import FeedId
from ffun.feeds_collections.collections import collections
from ffun.feeds_links import domain as fl_domain
from ffun.llms_framework import errors
from ffun.llms_framework.entities import (
    APIKeyUsage,
    KeyStatus,
    LLMApiKey,
    LLMConfiguration,
    LLMProvider,
    SelectKeyContext,
    USDCost,
    UserKeyInfo,
)
from ffun.llms_framework.provider_interface import ProviderInterface
from ffun.resources import domain as r_domain
from ffun.user_settings import domain as us_domain

logger = logging.get_module_logger()


# TODO: cost
class CostPoints:
    __slots__ = ("_k",)

    def __init__(self, k: int) -> None:
        self._k = k

    def to_cost(self, points: int) -> USDCost:
        return USDCost(Decimal(points) / self._k)

    def to_points(self, cost: USDCost) -> int:
        return int(cost * self._k)


_cost_points = CostPoints(k=1_000_000_000)


# Note: this code is not about billing, it is about protection from the overuse of keys
#       i.e. in case of a problem, we should count a key as used with a maximum used tokens


class KeySelector(Protocol):
    async def __call__(self, llm: ProviderInterface, context: SelectKeyContext) -> APIKeyUsage | None:
        """Selector function"""


#######################
# filtering users logic
#######################


# TODO: add lock here to not check the same key in parallel by different processors
async def _api_key_is_working(llm: ProviderInterface, llm_config: LLMConfiguration, api_key: str) -> bool:
    status = llm.api_keys_statuses.get(api_key)

    if status == KeyStatus.works:
        return True

    if status != KeyStatus.unknown:
        return False

    new_status = await llm.check_api_key(llm_config, api_key)

    return new_status == KeyStatus.works


async def _filter_out_users_with_wrong_keys(
    llm: ProviderInterface, llm_config: LLMConfiguration, infos: list[UserKeyInfo], **kwargs: Any
) -> list[UserKeyInfo]:
    return [info for info in infos if info.api_key and await _api_key_is_working(llm, llm_config, info.api_key)]


async def _filter_out_users_without_keys(infos: list[UserKeyInfo], **kwargs: Any) -> list[UserKeyInfo]:
    return [info for info in infos if info.api_key]


async def _filter_out_users_for_whome_entry_is_too_old(
    infos: list[UserKeyInfo], entry_age: datetime.timedelta, **kwargs: Any
) -> list[UserKeyInfo]:
    return [info for info in infos if info.process_entries_not_older_than >= entry_age]


async def _filter_out_users_with_overused_keys(
    infos: list[UserKeyInfo], reserved_cost: USDCost, **kwargs: Any
) -> list[UserKeyInfo]:
    return [info for info in infos if info.cost_used + reserved_cost < info.max_tokens_cost_in_month]


async def _choose_user(
    infos: list[UserKeyInfo], reserved_cost: USDCost, interval_started_at: datetime.datetime
) -> UserKeyInfo | None:
    from ffun.application.resources import Resource as AppResource

    for info in infos:

        if await r_domain.try_to_reserve(
            user_id=info.user_id,
            kind=AppResource.tokens_cost,
            interval_started_at=interval_started_at,
            amount=_cost_points.to_points(reserved_cost),
            limit=_cost_points.to_points(info.max_tokens_cost_in_month),
        ):
            return info

    return None


# TODO: test that works for openai and gemini
async def _get_user_key_infos(  # pylint: disable=R0914
    provider: LLMProvider, user_ids: Iterable[UserId], interval_started_at: datetime.datetime
) -> list[UserKeyInfo]:
    from ffun.application.resources import Resource as AppResource
    from ffun.application.user_settings import UserSetting

    # TODO: move somewhere in configs
    provider_to_settings = {
        LLMProvider.openai: UserSetting.openai_api_key,
        LLMProvider.google: UserSetting.gemini_api_key,
        LLMProvider.test: UserSetting.test_api_key,
    }

    key_setting = provider_to_settings[provider]

    kinds = [UserSetting.max_tokens_cost_in_month, UserSetting.process_entries_not_older_than, key_setting]

    users_settings = await us_domain.load_settings_for_users(
        user_ids,
        kinds=kinds,
    )

    resources = await r_domain.load_resources(
        user_ids=user_ids, kind=AppResource.tokens_cost, interval_started_at=interval_started_at
    )

    infos = []

    for user_id in user_ids:
        settings = users_settings[user_id]

        days = settings.get(UserSetting.process_entries_not_older_than)
        assert days is not None
        assert isinstance(days, int)

        max_tokens_cost_in_month_raw = settings.get(UserSetting.max_tokens_cost_in_month)
        assert isinstance(max_tokens_cost_in_month_raw, Decimal)
        max_tokens_cost_in_month = USDCost(max_tokens_cost_in_month_raw)

        api_key_raw = settings.get(key_setting)
        assert isinstance(api_key_raw, str)
        api_key = LLMApiKey(api_key_raw)

        infos.append(
            UserKeyInfo(
                user_id=user_id,
                api_key=api_key,
                max_tokens_cost_in_month=max_tokens_cost_in_month,
                process_entries_not_older_than=datetime.timedelta(days=days),
                cost_used=_cost_points.to_cost(resources[user_id].total),
            )
        )

    return infos


_filters = (
    _filter_out_users_without_keys,
    _filter_out_users_for_whome_entry_is_too_old,
    _filter_out_users_with_wrong_keys,
    _filter_out_users_with_overused_keys,
)


async def _get_candidates(  # noqa
    llm: ProviderInterface,
    llm_config: LLMConfiguration,
    feed_ids: set[FeedId],
    interval_started_at: datetime.datetime,
    entry_age: datetime.timedelta,
    reserved_cost: USDCost,
    filters: tuple[Any, ...] = _filters,
) -> list[UserKeyInfo]:
    user_ids = await fl_domain.get_linked_users_flat(feed_ids)

    infos = await _get_user_key_infos(llm.provider, user_ids, interval_started_at)

    for _filter in filters:

        if not infos:
            return []

        infos = await _filter(
            llm=llm, llm_config=llm_config, infos=infos, entry_age=entry_age, reserved_cost=reserved_cost
        )

    return infos


async def _find_best_user_with_key(
    llm: ProviderInterface,
    llm_config: LLMConfiguration,
    feed_ids: set[FeedId],
    entry_age: datetime.timedelta,
    interval_started_at: datetime.datetime,
    reserved_cost: USDCost,
) -> UserKeyInfo | None:
    infos = await _get_candidates(
        llm=llm,
        llm_config=llm_config,
        feed_ids=feed_ids,
        interval_started_at=interval_started_at,
        entry_age=entry_age,
        reserved_cost=reserved_cost,
    )

    infos.sort(key=lambda info: info.cost_used)

    return await _choose_user(infos=infos, reserved_cost=reserved_cost, interval_started_at=interval_started_at)


####################
# choosing key logic
####################


async def _choose_general_key(llm: ProviderInterface, context: SelectKeyContext) -> APIKeyUsage | None:

    key = context.general_api_key

    if key is None:
        return None

    return APIKeyUsage(
        provider=llm.provider,
        user_id=None,
        api_key=key,
        reserved_cost=context.reserved_cost,
        input_tokens=None,
        output_tokens=None,
        used_cost=None,
        interval_started_at=context.interval_started_at,
    )


async def _choose_collections_key(llm: ProviderInterface, context: SelectKeyContext) -> APIKeyUsage | None:

    key = context.collections_api_key

    if key is None:
        return None

    if all(not collections.has_feed(feed_id) for feed_id in context.feed_ids):
        return None

    return APIKeyUsage(
        provider=llm.provider,
        user_id=None,
        api_key=key,
        reserved_cost=context.reserved_cost,
        input_tokens=None,
        output_tokens=None,
        used_cost=None,
        interval_started_at=context.interval_started_at,
    )


async def _choose_user_key(llm: ProviderInterface, context: SelectKeyContext) -> APIKeyUsage | None:

    if any(collections.has_feed(feed_id) for feed_id in context.feed_ids):
        raise errors.FeedsFromCollectionsMustNotBeProcessedWithUserAPIKeys(feed_ids=context.feed_ids)

    info = await _find_best_user_with_key(
        llm=llm,
        llm_config=context.llm_config,
        feed_ids=context.feed_ids,
        entry_age=context.entry_age,
        interval_started_at=context.interval_started_at,
        reserved_cost=context.reserved_cost,
    )

    if info is None:
        return None

    if info.api_key is None:
        raise NotImplementedError("imporssible keys")

    return APIKeyUsage(
        provider=llm.provider,
        user_id=info.user_id,
        api_key=info.api_key,
        reserved_cost=context.reserved_cost,
        input_tokens=None,
        output_tokens=None,
        used_cost=None,
        interval_started_at=context.interval_started_at,
    )


# Collections key has priority over general key
# Because otherwise there will be not possible to use general key only for users
_key_selectors: Collection[KeySelector] = (_choose_collections_key, _choose_general_key, _choose_user_key)


async def choose_api_key(
    llm: ProviderInterface, context: SelectKeyContext, selectors: Iterable[KeySelector] = _key_selectors
) -> APIKeyUsage | None:
    for key_selector in selectors:
        key_usage = await key_selector(llm, context)

        if key_usage is not None:
            return key_usage

    return None


#################
# key usage logic
#################


@contextlib.asynccontextmanager
async def use_api_key(key_usage: APIKeyUsage) -> AsyncGenerator[None, None]:
    from ffun.application.resources import Resource as AppResource
    from ffun.llms_framework.providers import llm_providers

    log = logger.bind(function="_use_key", user_id=key_usage.user_id)

    try:
        yield

        if key_usage.used_cost is None:
            raise errors.UsedTokensHasNotSpecified()

        log.info("used_cost", used_cost=key_usage.used_cost)

    finally:
        log.info("convert_reserved_to_used", reserved_cost=key_usage.reserved_cost, used_cost=key_usage.used_cost)

        new_key_status = llm_providers.get(key_usage.provider).provider.api_keys_statuses.get(key_usage.api_key)

        log.business_event(  # type: ignore
            "llm_api_key_used",
            user_id=key_usage.user_id,
            llm_provider=key_usage.provider,
            new_key_status=new_key_status,
        )

        if key_usage.user_id is not None:
            await r_domain.convert_reserved_to_used(
                user_id=key_usage.user_id,
                kind=AppResource.tokens_cost,
                interval_started_at=key_usage.interval_started_at,
                used=_cost_points.to_points(key_usage.cost_to_register()),
                reserved=_cost_points.to_points(key_usage.reserved_cost),
            )

        log.info("resources_converted")



from ffun.domain.entities import FeedId
from ffun.feeds import errors, operations
from ffun.feeds.entities import Feed

save_feed = operations.save_feed
update_feed_info = operations.update_feed_info
get_next_feeds_to_load = operations.get_next_feeds_to_load
mark_feed_as_loaded = operations.mark_feed_as_loaded
mark_feed_as_failed = operations.mark_feed_as_failed
mark_feed_as_orphaned = operations.mark_feed_as_orphaned
get_feeds = operations.get_feeds
get_feed_ids_by_uids = operations.get_feed_ids_by_uids
get_source_ids = operations.get_source_ids
get_orphaned_feeds = operations.get_orphaned_feeds
tech_remove_feed = operations.tech_remove_feed
count_total_feeds = operations.count_total_feeds
count_total_feeds_per_state = operations.count_total_feeds_per_state
count_total_feeds_per_last_error = operations.count_total_feeds_per_last_error


async def save_feeds(feeds: list[Feed]) -> list[FeedId]:
    real_ids = []

    for feed in feeds:
        real_ids.append(await save_feed(feed))

    return real_ids


async def get_feed(feed_id: FeedId) -> Feed:
    feeds = await get_feeds([feed_id])

    if not feeds:
        raise errors.NoFeedFound(feed_id=feed_id)

    return feeds[0]



from typing import Any, Iterable

import psycopg

from ffun.core import logging
from ffun.core.postgresql import execute
from ffun.domain.domain import new_rule_id
from ffun.domain.entities import RuleId, UserId
from ffun.scores import errors
from ffun.scores.entities import Rule

logger = logging.get_module_logger()


def _normalize_tags(tags: Iterable[int]) -> list[int]:
    return list(sorted(tags))


def _key_from_tags(required_tags: Iterable[int], excluded_tags: Iterable[int]) -> str:
    return ",".join(map(str, required_tags)) + "|" + ",".join(map(str, excluded_tags))


def row_to_rule(row: dict[str, Any]) -> Rule:
    return Rule(
        id=row["id"],
        user_id=row["user_id"],
        required_tags=row["required_tags"],
        excluded_tags=row["excluded_tags"],
        score=row["score"],
        created_at=row["created_at"],
        updated_at=row["updated_at"],
    )


async def create_or_update_rule(
    user_id: UserId, required_tags: Iterable[int], excluded_tags: Iterable[int], score: int
) -> Rule:
    required_tags = set(required_tags)
    excluded_tags = set(excluded_tags)

    if required_tags & excluded_tags:
        raise errors.TagsIntersection()

    required_tags = _normalize_tags(required_tags)
    excluded_tags = _normalize_tags(excluded_tags)

    key = _key_from_tags(required_tags, excluded_tags)

    sql = """
        INSERT INTO s_rules (id, user_id, required_tags, excluded_tags, key, score)
        VALUES (%(id)s, %(user_id)s, %(required_tags)s, %(excluded_tags)s, %(key)s, %(score)s)
        RETURNING *
        """

    try:
        result = await execute(
            sql,
            {
                "id": new_rule_id(),
                "user_id": user_id,
                "required_tags": required_tags,
                "excluded_tags": excluded_tags,
                "key": key,
                "score": score,
            },
        )

        logger.business_event(
            "rule_created",
            user_id=user_id,
            rule_id=result[0]["id"],
            required_tags=required_tags,
            excluded_tags=excluded_tags,
            score=score,
        )
    except psycopg.errors.UniqueViolation:
        logger.info("rule_already_exists_change_score", key=key)

        sql = """
        UPDATE s_rules
        SET score = %(score)s, updated_at = NOW()
        WHERE user_id = %(user_id)s AND key = %(key)s
        RETURNING *
        """

        result = await execute(sql, {"user_id": user_id, "key": key, "score": score})

        logger.business_event(
            "rule_updated",
            user_id=user_id,
            rule_id=result[0]["id"],
            required_tags=required_tags,
            excluded_tags=excluded_tags,
            score=score,
        )

    return row_to_rule(result[0])


async def delete_rule(user_id: UserId, rule_id: RuleId) -> None:
    sql = """
        DELETE FROM s_rules
        WHERE user_id = %(user_id)s AND id = %(rule_id)s
        RETURNING id
        """

    result = await execute(sql, {"user_id": user_id, "rule_id": rule_id})

    if result:
        logger.business_event("rule_deleted", user_id=user_id, rule_id=rule_id)


async def update_rule(
    user_id: UserId, rule_id: RuleId, required_tags: Iterable[int], excluded_tags: Iterable[int], score: int
) -> Rule:
    required_tags = set(required_tags)
    excluded_tags = set(excluded_tags)

    if required_tags & excluded_tags:
        raise errors.TagsIntersection()

    required_tags = _normalize_tags(required_tags)
    excluded_tags = _normalize_tags(excluded_tags)

    key = _key_from_tags(required_tags, excluded_tags)

    sql = """
    UPDATE s_rules
    SET required_tags = %(required_tags)s,
        excluded_tags = %(excluded_tags)s,
        key = %(key)s,
        score = %(score)s,
        updated_at = NOW()
    WHERE user_id = %(user_id)s AND id = %(rule_id)s
    returning *
    """

    result = await execute(
        sql,
        {
            "user_id": user_id,
            "rule_id": rule_id,
            "required_tags": required_tags,
            "excluded_tags": excluded_tags,
            "key": key,
            "score": score,
        },
    )

    if not result:
        raise errors.NoRuleFound()

    logger.business_event(
        "rule_updated",
        user_id=user_id,
        rule_id=result[0]["id"],
        required_tags=required_tags,
        excluded_tags=excluded_tags,
        score=score,
    )

    return row_to_rule(result[0])


async def get_rules(user_id: UserId) -> list[Rule]:
    sql = """
    SELECT *
    FROM s_rules
    WHERE user_id = %(user_id)s
    """

    rows = await execute(sql, {"user_id": user_id})

    return [row_to_rule(row) for row in rows]


async def count_rules_per_user() -> dict[UserId, int]:
    # Not optimal implementation, but should work for a very long time
    result = await execute("SELECT user_id, COUNT(*) FROM s_rules GROUP BY user_id")

    return {row["user_id"]: row["count"] for row in result}



import datetime

from ffun.core import utils
from ffun.llms_framework.entities import KeyStatus
from ffun.llms_framework.settings import settings


class StatusInfo:
    __slots__ = ("status", "updated_at")

    def __init__(self, status: KeyStatus, updated_at: datetime.datetime) -> None:
        self.status = status
        self.updated_at = updated_at


class Statuses:
    __slots__ = ("_statuses",)

    def __init__(self) -> None:
        self._statuses: dict[str, StatusInfo] = {}

    def set(self, key: str, status: KeyStatus) -> None:
        self._statuses[key] = StatusInfo(status, utils.now())

    def get(self, key: str) -> KeyStatus:
        info = self._statuses.get(key)

        if info is None:
            return KeyStatus.unknown

        if info.status == KeyStatus.works:
            return KeyStatus.works

        if info.status == KeyStatus.broken and info.updated_at + settings.key_broken_timeout > utils.now():
            return KeyStatus.broken

        if info.status == KeyStatus.quota and info.updated_at + settings.key_quota_timeout > utils.now():
            return KeyStatus.quota

        return KeyStatus.unknown



from typing import Iterable

from ffun.core.postgresql import execute
from ffun.librarian import operations


async def clean_failed_storage(processor_ids: Iterable[int]) -> None:
    for processor_id in processor_ids:
        while True:
            failed_entries = await operations.get_failed_entries(execute, processor_id, limit=1000)
            await operations.remove_failed_entries(execute, processor_id, failed_entries)

            if not failed_entries:
                break



import httpx
from furl import furl

from ffun.core import logging, utils
from ffun.domain.domain import new_entry_id
from ffun.domain.entities import AbsoluteUrl, FeedUrl
from ffun.feeds import domain as f_domain
from ffun.feeds.entities import Feed, FeedId, FeedState
from ffun.feeds_collections.collections import collections
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.library import entities as l_entities
from ffun.loader import errors, operations
from ffun.loader.entities import ProxyState
from ffun.loader.settings import settings
from ffun.parsers import entities as p_entities

logger = logging.get_module_logger()


_user_agent: str = "unknown"


decode_content = operations.decode_content
parse_content = operations.parse_content


# TODO: tests
def initialize(user_agent: str) -> None:
    global _user_agent
    logger.info("initialize_loader", user_agent=user_agent)
    _user_agent = user_agent


# TODO: tests
async def load_content_with_proxies(url: FeedUrl) -> httpx.Response:  # noqa: CCR001
    url_object = furl(url)

    first_exception = None

    # TODO: cache states for some time
    proxy_states = await operations.get_proxy_states(names=[proxy.name for proxy in settings.proxies])

    if all(state == ProxyState.suspended for state in proxy_states.values()):
        logger.warning("all_proxies_suspended")
        raise errors.AllProxiesSuspended()

    # We try different protocols because users often make mistakes in the urls
    # to fix them we unite similar urls like http://example.com and https://example.com
    # => we need to check both protocols
    #
    # For now it is straightforward solution, but it should work
    # Most of the domains should support HTTPS => HTTP urls will not be used
    # but in case of full problem with url, we'll be doing unnecessary requests
    # and in case of HTTP-only urls we'll be doing unnecessary requests too
    #
    # ATTENTION: we iterate over protocols even if the user has specified the port, because
    #            1. We do not trust users.
    #            2. We check urls duplicates by removing ports (see domain.urls.url_to_uid
    #
    # TODO: build a separate system of choosing protocol for the url with caching and periodic checks
    for protocol in ("https", "http"):
        logger.info("try_protocol", protocol=protocol)

        url_object.scheme = protocol

        for proxy in settings.proxies:
            logger.info("try_proxy", proxy=proxy.name)

            if proxy_states[proxy.name] == ProxyState.suspended:
                logger.info("skip_suspended_proxy", proxy=proxy.name)
                continue

            try:
                return await operations.load_content(AbsoluteUrl(str(url_object)), proxy, _user_agent)
            except Exception as e:
                logger.info("proxy_error", proxy=proxy.name, error=e)

                if first_exception is None:
                    first_exception = e

        logger.info("all_proxies_failed", protocol=protocol)

    logger.info("all_protocols_failed")

    # in case of error raise the first exception occurred
    # because we should use the most common proxy first
    raise first_exception  # type: ignore


async def detect_orphaned(feed_id: FeedId) -> bool:
    if collections.has_feed(feed_id):
        return False

    if await fl_domain.has_linked_users(feed_id):
        return False

    logger.info("feed_has_no_linked_users")
    await f_domain.mark_feed_as_orphaned(feed_id)

    return True


# TODO: tests
async def extract_feed_info(feed_id: FeedId | None, feed_url: FeedUrl) -> p_entities.FeedInfo | None:
    try:
        response = await load_content_with_proxies(feed_url)
        content = await operations.decode_content(response)
        feed_info = await operations.parse_content(content, original_url=feed_url)
    except errors.AllProxiesSuspended:
        logger.info("all_proxies_suspended")
        return None
    except errors.LoadError as e:
        logger.info("feed_load_error", error_code=e.feed_error_code)

        if feed_id is not None:
            await f_domain.mark_feed_as_failed(feed_id, state=FeedState.damaged, error=e.feed_error_code)

        return None

    logger.info("feed_loaded", entries_number=len(feed_info.entries))

    return feed_info


async def sync_feed_info(feed: Feed, feed_info: p_entities.FeedInfo) -> None:
    title = feed_info.title
    description = feed_info.description

    if collections.has_feed(feed.id):
        collections_feed_info = collections.get_feed_info(feed.id)

        title = collections_feed_info.title
        description = collections_feed_info.description

    if title == feed.title and description == feed.description:
        return

    await f_domain.update_feed_info(feed.id, title=feed_info.title, description=feed_info.description)


async def store_entries(feed: Feed, entries: list[p_entities.EntryInfo]) -> None:
    external_ids = [entry.external_id for entry in entries]

    stored_entries_external_ids = await l_domain.find_stored_entries_for_feed(feed.id, external_ids)

    entries_to_store = [entry for entry in entries if entry.external_id not in stored_entries_external_ids]

    prepared_entries = [
        l_entities.Entry(
            id=new_entry_id(), source_id=feed.source_id, cataloged_at=utils.now(), **entry_info.model_dump()
        )
        for entry_info in entries_to_store
    ]

    await l_domain.catalog_entries(feed.id, entries=prepared_entries)

    entries_stored = len(prepared_entries)

    logger.info("entries_stored", entries_number=entries_stored)

    if entries_stored > 0:
        logger.business_event("news_entries_stored", user_id=None, feed_id=feed.id, entries_number=entries_stored)


@logging.function_args_to_log("feed.id", "feed.url")
async def process_feed(feed: Feed) -> None:
    logger.info("loading_feed")

    if await detect_orphaned(feed.id):
        return

    feed_info = await extract_feed_info(feed_id=feed.id, feed_url=feed.url)

    if feed_info is not None:
        await sync_feed_info(feed, feed_info)
        await store_entries(feed, feed_info.entries)
        await f_domain.mark_feed_as_loaded(feed.id)
        logger.info("entries_loaded")
    else:
        logger.info("feed_not_loaded")

    # We should sync the number of entries in the feed even if we failed to load it
    # to automatically sync the tail size after any configuration change.
    # We should not do so for orphaned feeds, because they will be removed.
    await l_domain.unlink_feed_tail(feed.id)


async def check_proxies_availability() -> None:
    states = {}

    for proxy in settings.proxies:
        is_available = await operations.is_proxy_available(
            proxy=proxy, anchors=settings.proxy_anchors, user_agent=_user_agent
        )

        states[proxy.name] = ProxyState.available if is_available else ProxyState.suspended

    await operations.update_proxy_states(states)



import datetime
import uuid

import pytest
import pytest_asyncio

from ffun.core import utils
from ffun.core.postgresql import execute
from ffun.core.tests.helpers import TableSizeDelta, TableSizeNotChanged
from ffun.domain.domain import new_entry_id
from ffun.domain.entities import EntryId
from ffun.feeds.entities import Feed
from ffun.librarian import errors, operations
from ffun.librarian.entities import ProcessorPointer
from ffun.librarian.tests import helpers
from ffun.library.tests import make as l_make


def assert_is_new_pointer(pointer: ProcessorPointer, processor_id: int) -> None:
    assert pointer.processor_id == processor_id
    assert pointer.pointer_created_at == datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    assert pointer.pointer_entry_id == EntryId(uuid.UUID("00000000-0000-0000-0000-000000000000"))


class TestCreatePointer:
    @pytest.mark.asyncio
    async def test_new_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeDelta("ln_processor_pointers", delta=1):
            pointer = await operations.create_pointer(fake_processor_id)

        assert pointer is not None

        assert_is_new_pointer(pointer, fake_processor_id)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer.processor_id == fake_processor_id

    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processor_pointers"):
            second_pointer = await operations.create_pointer(fake_processor_id)

        assert second_pointer is None

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert pointer == loaded_pointer


# the base behavior of the get_or_create_pointer function is checked in other tests, that use it
class TestGetOrCreatePointer:
    @pytest.mark.asyncio
    async def test_create_if_not_exists(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeDelta("ln_processor_pointers", delta=1):
            pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert_is_new_pointer(pointer, fake_processor_id)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert loaded_pointer.processor_id == fake_processor_id


class TestDeletePointer:
    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        pointer.replace(pointer_entry_id=uuid.uuid4())

        await operations.save_pointer(execute, pointer)

        async with TableSizeDelta("ln_processor_pointers", delta=-1):
            await operations.delete_pointer(fake_processor_id)

        new_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert_is_new_pointer(new_pointer, fake_processor_id)

    @pytest.mark.asyncio
    async def test_no_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        async with TableSizeNotChanged("ln_processor_pointers"):
            await operations.delete_pointer(fake_processor_id)


class TestSavePointer:
    @pytest.mark.asyncio
    async def test_no_pointer(self, fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)

        pointer = ProcessorPointer(
            processor_id=fake_processor_id,
            pointer_entry_id=new_entry_id(),
            pointer_created_at=datetime.datetime.now(),
        )

        async with TableSizeNotChanged("ln_processor_pointers"):
            with pytest.raises(errors.CanNotSaveUnexistingPointer):
                await operations.save_pointer(execute, pointer)

    @pytest.mark.asyncio
    async def test_existing_pointer(self, fake_processor_id: int) -> None:
        pointer = await operations.get_or_create_pointer(fake_processor_id)

        new_pointer = pointer.replace(pointer_created_at=utils.now(), pointer_entry_id=uuid.uuid4())

        async with TableSizeNotChanged("ln_processor_pointers"):
            await operations.save_pointer(execute, new_pointer)

        loaded_pointer = await operations.get_or_create_pointer(fake_processor_id)

        assert new_pointer == loaded_pointer


class TestGetAllPointers:

    @pytest_asyncio.fixture(autouse=True)
    async def cleanup_fake_pointers(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        await operations.delete_pointer(fake_processor_id)
        await operations.delete_pointer(another_fake_processor_id)

    @pytest.mark.asyncio
    async def test_no_pointers(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        pointers = await operations.get_all_pointers()

        ids = {pointer.processor_id for pointer in pointers}

        assert fake_processor_id not in ids
        assert another_fake_processor_id not in ids

    @pytest.mark.asyncio
    async def test_has_pointers(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        pointer_1 = await operations.get_or_create_pointer(fake_processor_id)
        pointer_2 = await operations.get_or_create_pointer(another_fake_processor_id)

        pointer_2 = pointer_2.replace(pointer_entry_id=new_entry_id(), pointer_created_at=utils.now())

        await operations.save_pointer(execute, pointer_2)

        pointers = await operations.get_all_pointers()

        ids = {pointer.processor_id: pointer for pointer in pointers}

        assert ids[fake_processor_id] == pointer_1
        assert ids[another_fake_processor_id] == pointer_2


class TestPushEntriesToProcessorQueue:
    @pytest.mark.parametrize("entries_count", [0, 1, 3, 10])
    @pytest.mark.asyncio
    async def test_push_entry(self, loaded_feed: Feed, entries_count: int, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=entries_count + 5)

        entries_to_push = list(entries)[:entries_count]

        async with TableSizeDelta("ln_processors_queue", delta=entries_count):
            await operations.push_entries_to_processor_queue(execute, fake_processor_id, entries_to_push)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=entries_count + 5)

        assert set(entries_to_push) == set(entries_in_queue)


class TestGetEntriesToProcess:
    @pytest.mark.asyncio
    async def test_get_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        received_entries = set()

        while True:
            new_entries_ids = await operations.get_entries_to_process(fake_processor_id, limit=3)
            received_entries.update(new_entries_ids)
            await operations.remove_entries_from_processor_queue(execute, fake_processor_id, new_entries_ids)

            if len(received_entries) == len(entries):
                assert len(new_entries_ids) == 13 % 3
                break

        assert set(entries) == received_entries

    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await operations.get_entries_to_process(fake_processor_id, limit=3)

        assert not entries

    @pytest.mark.asyncio
    async def test_idempotency(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_ids_1 = await operations.get_entries_to_process(fake_processor_id, limit=3)
        entries_ids_2 = await operations.get_entries_to_process(fake_processor_id, limit=3)

        assert entries_ids_1 == entries_ids_2


class TestCountEntriesInProcessorQueue:
    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        count = await operations.count_entries_in_processor_queue(fake_processor_id)

        assert count == 0

    @pytest.mark.asyncio
    async def test_count(self, loaded_feed: Feed, fake_processor_id: int, another_fake_processor_id: int) -> None:
        old_count = await operations.count_entries_in_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=3)
        more_entries = await l_make.n_entries(loaded_feed, n=2)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))
        await operations.push_entries_to_processor_queue(execute, another_fake_processor_id, list(more_entries))

        new_count = await operations.count_entries_in_processor_queue(fake_processor_id)

        assert new_count == old_count + 3


class TestRemoveEntriesFromProcessorQueue:
    @pytest.mark.asyncio
    async def test_remove_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_to_remove = list(entries)[:5]

        async with TableSizeDelta("ln_processors_queue", delta=-5):
            await operations.remove_entries_from_processor_queue(execute, fake_processor_id, entries_to_remove)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=13)

        assert len(entries_in_queue) == 8

        assert set(entries_to_remove) & set(entries_in_queue) == set()

    @pytest.mark.asyncio
    async def test_remove_non_existing_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        await operations.clear_processor_queue(fake_processor_id)

        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.push_entries_to_processor_queue(execute, fake_processor_id, list(entries))

        entries_to_remove = [new_entry_id(), new_entry_id()]

        await operations.remove_entries_from_processor_queue(execute, fake_processor_id, entries_to_remove)

        entries_in_queue = await operations.get_entries_to_process(fake_processor_id, limit=13)

        assert set(entries) == set(entries_in_queue)


class TestClearProcessorQueue:
    # checked in other tests
    pass


class TestAddEntriesToFailedStorage:
    @pytest.mark.asyncio
    async def test_add_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed, n=13)

        entries_to_add = list(entries)[:5]

        async with TableSizeDelta("ln_failed_entries", delta=5):
            await operations.add_entries_to_failed_storage(fake_processor_id, entries_to_add)

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries_to_add) <= set(failed_entries)


class TestGetFailedEntries:
    @pytest.mark.asyncio
    async def test_get_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed, n=4)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries) <= set(failed_entries)

    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int) -> None:
        while True:
            failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=1000)
            await operations.remove_failed_entries(execute, fake_processor_id, failed_entries)

            if not failed_entries:
                break

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert not failed_entries

    @pytest.mark.asyncio
    async def test_idempotency(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        failed_entries_1 = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)
        failed_entries_2 = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert failed_entries_1 == failed_entries_2


class TestRemoveFailedEntries:
    @pytest.mark.asyncio
    async def test_remove_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed, n=13)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        entries_to_remove = list(entries)[:5]

        async with TableSizeDelta("ln_failed_entries", delta=-5):
            await operations.remove_failed_entries(execute, fake_processor_id, entries_to_remove)

        failed_entries = await operations.get_failed_entries(execute, fake_processor_id, limit=100500)

        assert set(entries_to_remove) & set(failed_entries) == set()

    @pytest.mark.asyncio
    async def test_remove_non_existing_entries(self, loaded_feed: Feed, fake_processor_id: int) -> None:
        entries = await l_make.n_entries(loaded_feed, n=5)

        await operations.add_entries_to_failed_storage(fake_processor_id, list(entries))

        entries_to_remove = [new_entry_id(), new_entry_id()]

        async with TableSizeNotChanged("ln_failed_entries"):
            await operations.remove_failed_entries(execute, fake_processor_id, entries_to_remove)


class TestCountFailedEntries:
    @pytest.mark.asyncio
    async def test_no_entries(self, fake_processor_id: int, another_fake_processor_id: int) -> None:
        await helpers.clean_failed_storage([fake_processor_id, another_fake_processor_id])

        counts = await operations.count_failed_entries()

        assert fake_processor_id not in counts
        assert another_fake_processor_id not in counts

    @pytest.mark.asyncio
    async def test_some_entries(
        self, saved_feed: Feed, fake_processor_id: int, another_fake_processor_id: int
    ) -> None:
        await helpers.clean_failed_storage([fake_processor_id, another_fake_processor_id])

        entries = await l_make.n_entries(saved_feed, n=3)
        entries_list = list(entries)

        await operations.add_entries_to_failed_storage(fake_processor_id, [entries_list[0], entries_list[1]])
        await operations.add_entries_to_failed_storage(
            another_fake_processor_id, [entry_id for entry_id in entries_list]
        )

        counts = await operations.count_failed_entries()

        assert counts[fake_processor_id] == 2
        assert counts[another_fake_processor_id] == 3



import pytest

from ffun.librarian.processors import upper_case_title
from ffun.library.entities import Entry
from ffun.ontology.entities import ProcessorTag

processor = upper_case_title.Processor(name="upper_case_title")

expected_tag = "upper-case-title"


class TestEncodeSpecialCharacters:
    @pytest.mark.parametrize(
        "title, has_tag",
        [
            ("", False),
            ("UPPER CASE", True),
            ("  UPPER ,CASE  ", True),
            ("«СЛОЖНЫЙ ЮНИКОД» #324", True),
            ("«Сложный Юникод» #324", False),
            ("Normal title?", False),
            ("lower case only title", False),
            ("только нижний кейс юникод", False),
        ],
    )
    @pytest.mark.asyncio
    async def test(self, cataloged_entry: Entry, title: str, has_tag: bool) -> None:
        cataloged_entry = cataloged_entry.replace(title=title)

        tags = await processor.process(cataloged_entry)

        expected_tags = []

        if has_tag:
            expected_tags.append(ProcessorTag(raw_uid=expected_tag))

        assert tags == expected_tags



from ffun.core.postgresql import ExecuteType, run_in_transaction
from ffun.domain.entities import EntryId
from ffun.markers import operations

set_marker = operations.set_marker
remove_marker = operations.remove_marker
get_markers = operations.get_markers
remove_markers_for_entries = operations.remove_markers_for_entries


@run_in_transaction
async def tech_merge_markers(execute: ExecuteType, from_entry_id: EntryId, to_entry_id: EntryId) -> None:
    await operations.tech_merge_markers(execute, from_entry_id, to_entry_id)



from typing import Iterable

from ffun.core import logging, utils
from ffun.domain.domain import new_feed_id
from ffun.domain.entities import EntryId, FeedId, UserId
from ffun.domain.urls import url_to_source_uid
from ffun.feeds import domain as f_domain
from ffun.feeds import entities as f_entities
from ffun.feeds_links import domain as fl_domain
from ffun.library import domain as l_domain
from ffun.markers import domain as m_domain
from ffun.meta.settings import settings
from ffun.ontology import domain as o_domain
from ffun.parsers import entities as p_entities

logger = logging.get_module_logger()


# Note: fully unlinked entry can be linked again before removing from l_entries
#       we should do something with it in the future, but it is ok for now
#       there should not a lot of such cases
async def remove_entries(entries_ids: Iterable[EntryId]) -> None:
    """Remove entries and all related markers and relations."""
    entries_to_remove = list(entries_ids)

    await m_domain.remove_markers_for_entries(entries_to_remove)
    await o_domain.remove_relations_for_entries(entries_to_remove)
    await l_domain.remove_entries_by_ids(entries_to_remove)


async def add_feeds(feed_infos: list[p_entities.FeedInfo], user_id: UserId) -> list[FeedId]:

    urls_to_sources_uids = {feed_info.url: url_to_source_uid(feed_info.url) for feed_info in feed_infos}

    sources_uids_to_ids = await f_domain.get_source_ids(urls_to_sources_uids.values())

    feeds = [
        f_entities.Feed(
            id=new_feed_id(),
            source_id=sources_uids_to_ids[urls_to_sources_uids[feed_info.url]],
            url=feed_info.url,
            title=feed_info.title,
            description=feed_info.description,
        )
        for feed_info in feed_infos
    ]

    real_feeds_ids = await f_domain.save_feeds(feeds)

    for feed_id in real_feeds_ids:
        await fl_domain.add_link(user_id=user_id, feed_id=feed_id)

    return real_feeds_ids


async def clean_orphaned_entries(chunk: int) -> int:
    orphanes = await l_domain.get_orphaned_entries(limit=chunk)

    await remove_entries(orphanes)

    return len(orphanes)


# There is a very small possibility, that user will link feed
# while we in the process of removing it as an orphaned feed.
# This is look like an almost impossible situation (on the current load), because:
# - the feed must be marked as orphaned before that
# - we wait a timeout before removing orphaned feeds
# - the user should make their operations right at the time we process orphaned feeds
async def clean_orphaned_feeds(chunk: int) -> int:
    loaded_before = utils.now() - settings.delay_before_removing_orphaned_feeds

    orphanes = await f_domain.get_orphaned_feeds(limit=chunk, loaded_before=loaded_before)

    # ensure deterministic order of processing
    orphanes.sort()

    # refactor this loop into a bulk operation calls in case of performance issues
    for orphan_id in orphanes:
        logger.info("removing_orphaned_feed", feed_id=orphan_id)

        # unlink all linked entries
        await l_domain.unlink_feed_tail(orphan_id, offset=0)

        await f_domain.tech_remove_feed(orphan_id)

    # just a protection in case some user linked feed while we were removing it
    # in that case return DB in the consistent state
    await fl_domain.tech_remove_all_links(orphanes)

    return len(orphanes)



from ffun.core import errors


class Error(errors.Error):
    pass


class ModelDoesNotFound(Error):
    pass


class TextPartsMustBePositive(Error):
    pass


class TextIntersectionMustBePositiveOrZero(Error):
    pass


class TextIsEmpty(Error):
    pass


class TextIsTooShort(Error):
    pass


class TemporaryError(Error):
    pass


class UsedTokensHasNotSpecified(Error):
    pass


class FeedsFromCollectionsMustNotBeProcessedWithUserAPIKeys(Error):
    pass
